#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import functools
import logging
import re
from typing import List, Mapping, Sequence

from flask import current_app, request, jsonify

from label_sleuth.analysis_utils.analyze_tokens import ngrams_by_info_gain
from label_sleuth.data_access.core.data_structs import TextElement
from label_sleuth.data_access.data_access_api import get_document_uri
from label_sleuth.models.core.languages import Languages
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import Iteration, IterationStatus
from label_sleuth.orchestrator.orchestrator_api import TRAIN_COUNTS_STR_KEY


def validate_workspace_id(function):
    @functools.wraps(function)
    def wrapper(workspace_id, *args, **kwargs):
        if not current_app.orchestrator_api.workspace_exists(workspace_id):
            return jsonify({"type": "workspace_id_does_not_exist",
                            "title": f"workspace_id {workspace_id} does not exist"}), 404
        return function(workspace_id, *args, **kwargs)

    return wrapper


def validate_category_id(function):
    @functools.wraps(function)
    def wrapper(workspace_id, *args, **kwargs):
        category_id = request.args.get('category_id')
        if category_id is None:
            return jsonify({"type": "missing_category_id", "title": "category_id was not provided"}), 422

        category_id = int(category_id)
        if category_id not in current_app.orchestrator_api.get_all_categories(workspace_id):
            return jsonify({"type": "category_id_does_not_exist",
                            "title": f"category_id {category_id} does not exist in workspace {workspace_id}"}), 404
        return function(workspace_id, *args, **kwargs)

    return wrapper


def elements_back_to_front(workspace_id: str, elements: List[TextElement], category_id: int) -> List[Mapping]:
    """
    Converts TextElement objects from the backend into dictionaries in the form expected by the frontend, and adds
    the model prediction for the elements if available.
    :param workspace_id:
    :param elements: a list of TextElements
    :param category_id:
    :return: a list of dictionaries with element information
    """

    element_uri_to_info = \
        {text_element.uri:
             {'id': text_element.uri,
              'docid': get_document_uri(text_element.uri),
              'begin': text_element.span[0][0],
              'end': text_element.span[0][1],
              'text': text_element.text,
              'user_labels': {k: str(v.label).lower()
                              # TODO current UI is using true and false as strings. change to boolean in the new UI
                              for k, v in text_element.category_to_label.items()},
              'model_predictions': {}
              }
         for text_element in elements}

    if category_id is not None and len(elements) > 0 \
            and len(current_app.orchestrator_api.get_all_iterations_by_status(workspace_id, category_id,
                                                                              IterationStatus.READY)) > 0:
        predicted_labels = [pred.label
                            for pred in current_app.orchestrator_api.infer(workspace_id, category_id, elements)]
        for text_element, prediction in zip(elements, predicted_labels):
            # the frontend expects string labels and not boolean
            element_uri_to_info[text_element.uri]['model_predictions'][category_id] = str(prediction).lower()

    return [element_info for element_info in element_uri_to_info.values()]


def get_element(workspace_id, category_id, element_id):
    """
    Get element by id
    :param workspace_id:
    :param category_id:
    :param element_id:
    """
    dataset_name = current_app.orchestrator_api.get_dataset_name(workspace_id)
    element = current_app.orchestrator_api.get_text_elements_by_uris(workspace_id, dataset_name, [element_id])
    element_transformed = elements_back_to_front(workspace_id, element, category_id)[0]
    return element_transformed


def _extract_model_info(iteration):
    # TODO reorganize model information
    placeholder = 'Unknown'
    model_info_dict = {'model_status': placeholder, 'creation_epoch': placeholder, 'model_type': placeholder,
                       'model_metadata': placeholder}
    if iteration.model is not None:
        model_info_dict['model_status'] = iteration.model.model_status.name
        model_info_dict['creation_epoch'] = iteration.model.creation_date.timestamp()
        model_info_dict['model_type'] = iteration.model.model_type.name \
            if iteration.model.model_type is not None else placeholder
        # The frontend expects a dict of string to int, and train counts contains a mix of boolean and string keys.
        model_info_dict['model_metadata'] = \
            {**iteration.model.train_statistics,
             TRAIN_COUNTS_STR_KEY: {str(k).lower(): v for k, v
                                    in iteration.model.train_statistics.get(TRAIN_COUNTS_STR_KEY, {}).items()}}
    return model_info_dict


def extract_iteration_information_list(iterations: Sequence[Iteration]):
    return \
        [{'iteration': iteration_index,
          **iteration.iteration_statistics,
          'iteration_status': iteration.status.name,
          **_extract_model_info(iteration)}
         for iteration_index, iteration in enumerate(iterations)]


def extract_enriched_ngrams_and_weights_list(elements, boolean_labels):
    try:
        texts = [element.text for element in elements]
        enriched_ngrams_and_weights = ngrams_by_info_gain(texts, boolean_labels, ngram_max_length=2,
                                                          language=Languages.ENGLISH)
    except Exception as e:
        logging.warning(f"Failed to calculate enriched tokens from {len(elements)} elements: error {e}")
        enriched_ngrams_and_weights = {}

    ngrams_and_weights_list = [{'text': ngram, 'weight': weight} for ngram, weight in enriched_ngrams_and_weights]
    return ngrams_and_weights_list[:30]


def get_natural_sort_key(text):
    return [int(x) if x.isdigit() else x for x in re.split(r'(\d+)', text)]



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import io
import logging
import os
import time
import tempfile
import unittest
from label_sleuth import app, config
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import IterationStatus

HEADERS = {'Content-Type': 'application/json'}


class TestAppIntegration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()
        path_to_test_config = os.path.abspath(os.path.join(__file__, os.pardir, "config_for_tests.json"))
        loaded_config = config.load_config(path_to_test_config)
        # avoid downloading spacy model for test
        loaded_config.language.spacy_model_name = None
        app_for_test = app.create_app(config=loaded_config,
                                      output_dir=os.path.join(cls.temp_dir.name, 'output'))
        app_for_test.test_request_context("/")
        app_for_test.config['TESTING'] = True
        app_for_test.config['LOGIN_DISABLED'] = True

        print(f"Integration tests, all output files will be written under {cls.temp_dir}")
        cls.client = app_for_test.test_client()

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()

    def test_full_flow(self):
        self.maxDiff = None
        dataset_name = "my_test_dataset"
        workspace_name = "my_test_workspace"
        category_name = "my_category"
        category_description = "my_category_description"
        data = {}
        data['file'] = (io.BytesIO(
            b'document_id,text\n'
            b'document1,this is the first text element of document one\n'
            b'document1,this is the second text element of document one\n'
            b'document2,this is the only text element in document two\n'
            b'document3,"document 3 has three text elements, this is the first"\n'
            b'document3,"document 3 has three text elements, this is the second that will be labeled as negative"\n'
            b'document3,"document 3 has three text elements, this is the third"\n'),
                        'my_file.csv')
        res = self.client.post(f"/datasets/{dataset_name}/add_documents", data=data, headers=HEADERS,
                               content_type='multipart/form-data')

        self.assertEqual(200, res.status_code, msg="Failed to upload a new dataset")
        self.assertEqual(
            {'dataset_name': 'my_test_dataset', 'num_docs': 3, 'num_sentences': 6, 'workspaces_to_update': []},
            res.get_json(), msg="diff in upload dataset response")
        res = self.client.post("/workspace",
                               data='{{"workspace_id":"{}","dataset_id":"{}"}}'.format(workspace_name, dataset_name),
                               headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to create a workspace")
        self.assertEqual(
            {"workspace": {'dataset_name': 'my_test_dataset', 'first_document_id': 'my_test_dataset-document1',
                           'workspace_id': 'my_test_workspace'}}, res.get_json(),
            msg="diff in create workspace response")
        res = self.client.post(f"/workspace/{workspace_name}/category",
                               data='{{"category_name":"{}","category_description":"{}"}}'.format(category_name,
                                                                                                  category_description),
                               headers=HEADERS)
        category_id = int(res.get_json()["category_id"])
        self.assertEqual(200, res.status_code, msg="Failed to add a new category to workspace")
        self.assertEqual({'category_description': 'my_category_description',
                          'category_name': category_name, 'category_id': str(category_id)}, res.get_json(),
                         msg="diff in create category response")

        res = self.client.get(f"/workspace/{workspace_name}/documents", headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get all documents uris")
        documents = res.get_json()['documents']
        self.assertEqual(3, len(documents),
                         msg="Number of retrieved documents is different the number of documents loaded")
        self.assertEqual({'documents': [{'document_id': 'my_test_dataset-document1'},
                                        {'document_id': 'my_test_dataset-document2'},
                                        {'document_id': 'my_test_dataset-document3'}]}, res.get_json(),
                         msg="diff in get documents")
        res = self.client.get(f"/workspace/{workspace_name}/document/{documents[-1]['document_id']}", headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to add a document before labeling")
        document3_elements = res.get_json()['elements']
        self.assertEqual([
            {'begin': 0, 'docid': 'my_test_dataset-document3', 'end': 53, 'id': 'my_test_dataset-document3-0',
             'model_predictions': {}, 'text': 'document 3 has three text elements, this is the first',
             'user_labels': {}},
            {'begin': 54, 'docid': 'my_test_dataset-document3', 'end': 141, 'id': 'my_test_dataset-document3-1',
             'model_predictions': {},
             'text': 'document 3 has three text elements, this is the second that will be labeled as negative',
             'user_labels': {}},
            {'begin': 142, 'docid': 'my_test_dataset-document3', 'end': 195, 'id': 'my_test_dataset-document3-2',
             'model_predictions': {}, 'text': 'document 3 has three text elements, this is the third',
             'user_labels': {}}], document3_elements, msg=f"diff in {documents[-1]['document_id']} content")

        res = self.client.put(f'/workspace/{workspace_name}/element/{document3_elements[0]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to set the first label for a category")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 0, 'docid': 'my_test_dataset-document3', 'end': 53,
                                      'id': 'my_test_dataset-document3-0', 'model_predictions': {},
                                      'text': 'document 3 has three text elements, this is the first',
                                      'user_labels': {str(category_id): 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json(), msg="diff in setting element's label response")
        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get status after successfully setting the first label")
        self.assertEqual({'labeling_counts': {'true': 1}, 'progress': {'all': 50}},
                         res.get_json(), msg="diffs in get status response after setting a label")

        res = self.client.put(f'/workspace/{workspace_name}/element/{document3_elements[1]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, False),
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to set the second label for a category")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 54, 'docid': 'my_test_dataset-document3', 'end': 141,
                                      'id': 'my_test_dataset-document3-1', 'model_predictions': {},
                                      'text': 'document 3 has three text elements, '
                                              'this is the second that will be labeled as negative',
                                      'user_labels': {str(category_id): 'false'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json(), msg="diff in setting element's label response")
        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get status after successfully setting the first label")
        self.assertEqual({'labeling_counts': {'true': 1, 'false': 1}, 'progress': {'all': 50}},
                         res.get_json(), msg="diffs in get status response after setting a label")

        res = self.client.put(f'/workspace/{workspace_name}/element/{document3_elements[2]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)

        self.assertEqual(200, res.status_code, msg="Failed to set the third label for category")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 142, 'docid': 'my_test_dataset-document3', 'end': 195,
                                      'id': 'my_test_dataset-document3-2', 'model_predictions': {},
                                      'text': 'document 3 has three text elements, this is the third',
                                      'user_labels': {str(category_id): 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get status after successfully setting the third label")
        self.assertEqual({'true': 2, 'false': 1},
                         res.get_json()['labeling_counts'],
                         msg="diffs in get status response after setting the second label")

        res = self.wait_for_new_iteration(category_id, res, workspace_name, 1)

        self.assertEqual(200, res.status_code, msg="Failed to get iterations list")
        self.assertEqual(1, len(res.get_json()["iterations"]), msg="first model was not added to the models list")

        # get active learning recommendations
        res = self.client.get(f"/workspace/{workspace_name}/active_learning?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get active learning recommendations")
        active_learning_response = res.get_json()
        self.assertEqual({'elements': [
            {'begin': 47, 'docid': 'my_test_dataset-document1', 'end': 94, 'id': 'my_test_dataset-document1-1',
             'model_predictions': {str(category_id): 'true'}, 'text': 'this is the second text element of document one',
             'user_labels': {}},
            {'begin': 0, 'docid': 'my_test_dataset-document2', 'end': 45, 'id': 'my_test_dataset-document2-0',
             'model_predictions': {str(category_id): 'true'}, 'text': 'this is the only text element in document two',
             'user_labels': {}},
            {'begin': 0, 'docid': 'my_test_dataset-document1', 'end': 46, 'id': 'my_test_dataset-document1-0',
             'model_predictions': {str(category_id): 'true'}, 'text': 'this is the first text element of document one',
             'user_labels': {}}], 'hit_count': 3},
            active_learning_response)

        # set the first label according to the active learning recommendations
        res = self.client.put(f'/workspace/{workspace_name}/element/{active_learning_response["elements"][0]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)

        self.assertEqual(200, res.status_code,
                         msg="Failed to set the label for the first element recommended by the active learning")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 47, 'docid': 'my_test_dataset-document1', 'end': 94,
                                      'id': 'my_test_dataset-document1-1',
                                      'model_predictions': {str(category_id): 'true'},
                                      'text': 'this is the second text element of document one',
                                      'user_labels': {str(category_id): 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get status after successfully setting the first label for the second model")
        self.assertEqual({'labeling_counts': {'true': 3, 'false': 1}, 'progress': {'all': 50}},
                         res.get_json(), msg="diffs in get status response after setting a label")

        # set the second label according to the active learning recommendations
        res = self.client.put(f'/workspace/{workspace_name}/element/{active_learning_response["elements"][1]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, False),
                              headers=HEADERS)

        self.assertEqual(200, res.status_code,
                         msg="Failed to set the label for the first element recommended by the active learning")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 0, 'docid': 'my_test_dataset-document2', 'end': 45,
                                      'id': 'my_test_dataset-document2-0',
                                      'model_predictions': {str(category_id): 'true'},
                                      'text': 'this is the only text element in document two',
                                      'user_labels': {str(category_id): 'false'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get status after successfully setting the first label for the second model")
        self.assertEqual({'true': 3, 'false': 2},
                         res.get_json()['labeling_counts'], msg="diffs in get status response after setting a label")

        # get positively labeled elements
        res = self.client.get(f"/workspace/{workspace_name}/positive_elements?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get positively labeled elements")
        self.assertEqual({'elements': [
                            {'begin': 0,
                            'docid': 'my_test_dataset-document3',
                            'end': 53,
                            'id': 'my_test_dataset-document3-0',
                            'model_predictions': {'0': 'true'},
                            'text': 'document 3 has three text elements, this is '
                                    'the first',
                            'user_labels': {'0': 'true'}},
                            {'begin': 47,
                            'docid': 'my_test_dataset-document1',
                            'end': 94,
                            'id': 'my_test_dataset-document1-1',
                            'model_predictions': {'0': 'true'},
                            'text': 'this is the second text element of document '
                                    'one',
                            'user_labels': {'0': 'true'}},
                            {'begin': 142,
                            'docid': 'my_test_dataset-document3',
                            'end': 195,
                            'id': 'my_test_dataset-document3-2',
                            'model_predictions': {'0': 'true'},
                            'text': 'document 3 has three text elements, this is '
                                    'the third',
                            'user_labels': {'0': 'true'}}], 
                        'hit_count': 3},
                        res.get_json(), msg="diffs in positively labeled elements")

        # get negatively labeled elements
        res = self.client.get(f"/workspace/{workspace_name}/negative_elements?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get negatively labeled elements")
        self.assertEqual({'elements': [
                            {'begin': 0,
                            'docid': 'my_test_dataset-document2',
                            'end': 45,
                            'id': 'my_test_dataset-document2-0',
                            'model_predictions': {'0': 'true'},
                            'text': 'this is the only text element in document two',
                            'user_labels': {'0': 'false'}},
                            {'begin': 54,
                            'docid': 'my_test_dataset-document3',
                            'end': 141,
                            'id': 'my_test_dataset-document3-1',
                            'model_predictions': {'0': 'false'},
                            'text': 'document 3 has three text elements, this is '
                                    'the second that will be labeled as negative',
                            'user_labels': {'0': 'false'}}],
                        'hit_count': 2},
                        res.get_json(), msg="diffs in negatively labeled elements")

        # wait for the second models
        res = self.wait_for_new_iteration(category_id, res, workspace_name, 2)
        self.assertEqual(200, res.status_code, msg="Failed to get models list")
        self.assertEqual(2, len(res.get_json()["iterations"]), msg="second model was not added to the models list")

        # update existing category
        new_category_name = f'{category_name}_new'
        new_category_description = f'{category_description} new'
        res = self.client.put(f"/workspace/{workspace_name}/category/{category_id}",
                              data='{{"category_name":"{}","category_description":"{}"}}'.format(
                                  new_category_name, new_category_description),
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to update category")

        # delete category and create a new one
        res = self.client.delete(f"/workspace/{workspace_name}/category/{category_id}", headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to delete category")

        res = self.client.get(f"/workspace/{workspace_name}/categories", headers=HEADERS)
        self.assertEqual(len(res.get_json()['categories']), 0, msg='unexpected number of categories')

        res = self.client.post(f"/workspace/{workspace_name}/category",
                               data='{{"category_name":"{}","category_description":"{}"}}'.format(category_name,
                                                                                                  category_description),
                               headers=HEADERS)
        category_id = int(res.get_json()["category_id"])
        self.assertEqual(200, res.status_code, msg="Failed to add a new category to workspace")
        self.assertEqual(res.get_json(),
                         {'category_description': category_description, 'category_name': category_name,
                          'category_id': str(category_id)}, msg="diff in create category response")

        res = self.client.get(f"/workspace/{workspace_name}/categories", headers=HEADERS)
        self.assertEqual(len(res.get_json()['categories']), 1, msg='unexpected number of categories')

        print("Done")

    def wait_for_new_iteration(self, category_id, res, workspace_name, num_models):
        waiting_count = 0
        MAX_WAITING_FOR_TRAINING = 100
        # wait maximum 10 seconds (100*0.1) for the training
        # (should be much faster but it sometimes takes longer on github actions)
        while waiting_count < MAX_WAITING_FOR_TRAINING:
            # since get_status is asynchronously starting a new training, we need to wait until it added to the
            # iterations list and finishes successfully
            res = self.client.get(f"/workspace/{workspace_name}/iterations?category_id={category_id}",
                                  headers=HEADERS)
            response = res.get_json()
            print(response)
            if res.status_code != 200 or (len(response["iterations"]) == num_models
                                          and response['iterations'][0]['iteration_status'] == IterationStatus.READY.name):
                break
            time.sleep(0.1)
            waiting_count += 1
        return res



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import random
import unittest

from label_sleuth.active_learning.strategies.hard_example_mining import HardMiningLearner
from label_sleuth.active_learning.strategies.hybrid_learner import HybridLearner
from label_sleuth.active_learning.strategies.random_sampling import RandomSampling
from label_sleuth.active_learning.strategies.retrospective import RetrospectiveLearner

from label_sleuth.data_access.file_based.utils import URI_SEP
from label_sleuth.data_access.core.data_structs import Document, TextElement, Label, LABEL_NEGATIVE
from label_sleuth.models.core.prediction import Prediction


def generate_simple_doc(dataset_name, category_id, num_sentences, doc_id=0):
    sentence_template = 'This is sentence number :{}'
    sentences = [sentence_template.format(idx) for idx in range(num_sentences)]

    text_elements = []
    start_span = 0
    for idx, sentence in enumerate(sentences):
        end_span = start_span + len(sentence)
        mock_label = {category_id: Label(LABEL_NEGATIVE)}
        text_elements.append(TextElement(uri=URI_SEP.join([dataset_name, str(doc_id), str(idx)]), text=sentence,
                                         span=[(start_span, end_span)], metadata={}, category_to_label=mock_label))
        start_span = end_span + 1

    doc = Document(uri=dataset_name + URI_SEP + str(doc_id), text_elements=text_elements, metadata={})
    return doc


class TestActiveLearningStrategies(unittest.TestCase):
    def test_random_reproducibility(self):
        al = RandomSampling()
        category_id = 0
        doc = generate_simple_doc("dummy_dataset", category_id, num_sentences=100)
        predictions = [Prediction(True, random.random()) for _ in doc.text_elements]
        sorted_items_for_labeling1 = al.get_recommended_items_for_labeling("dummy_workspace", "dummy_dataset",
                                                                           category_id, doc.text_elements,
                                                                           predictions, sample_size=100)
        sorted_items_for_labeling2 = al.get_recommended_items_for_labeling("dummy_workspace", "dummy_dataset",
                                                                           category_id, doc.text_elements,
                                                                           predictions, sample_size=100)
        self.assertEqual(sorted_items_for_labeling1, sorted_items_for_labeling2)

    def test_hard_mining(self):
        al = HardMiningLearner()
        category_id = 0
        doc = generate_simple_doc("dummy_dataset", category_id, num_sentences=5)
        predictions = [Prediction(True, 1), Prediction(True, 0),
                       Prediction(True, 0.51), Prediction(True, 0.5001), Prediction(True, 0.52)]
        sorted_items_for_labeling = al.get_recommended_items_for_labeling("dummy_workspace", "dummy_dataset",
                                                                          category_id, doc.text_elements,
                                                                          predictions, sample_size=2)

        self.assertEqual(doc.text_elements[3], sorted_items_for_labeling[0])
        self.assertEqual(doc.text_elements[2], sorted_items_for_labeling[1])

    def test_retrospective(self):
        al = RetrospectiveLearner()
        category_id = 0
        doc = generate_simple_doc("dummy_dataset", category_id, num_sentences=5)
        predictions = [Prediction(True, 0.56), Prediction(True, 0),
                       Prediction(True, 0.99), Prediction(True, 1), Prediction(True, 0.52)]
        sorted_items_for_labeling = al.get_recommended_items_for_labeling("dummy_workspace", "dummy_dataset",
                                                                          category_id, doc.text_elements,
                                                                          predictions, sample_size=2)

        self.assertEqual(doc.text_elements[3], sorted_items_for_labeling[0])
        self.assertEqual(doc.text_elements[2], sorted_items_for_labeling[1])

    def test_hybrid_learner(self):
        al = HybridLearner(HardMiningLearner(), RetrospectiveLearner())
        category_id = 0
        doc = generate_simple_doc("dummy_dataset", category_id, num_sentences=5)
        predictions = [Prediction(True, 0), Prediction(True, 0.2),
                       Prediction(True, 0.5), Prediction(True, 0.75), Prediction(True, 1)]
        sorted_items_for_labeling = al.get_recommended_items_for_labeling("dummy_workspace", "dummy_dataset",
                                                                          category_id, doc.text_elements,
                                                                          predictions, sample_size=2)
        self.assertEqual(doc.text_elements[2], sorted_items_for_labeling[0])
        self.assertEqual(doc.text_elements[3], sorted_items_for_labeling[1])



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from dataclasses import dataclass
from typing import List

from datasets import Dataset
from tqdm.auto import tqdm
from transformers import AutoModelForSequenceClassification, AutoTokenizer, InputFeatures, Trainer, TrainingArguments, \
    TextClassificationPipeline, PreTrainedModel
from transformers.pipelines.pt_utils import KeyDataset

from label_sleuth.models.core.languages import Language, Languages
from label_sleuth.definitions import GPU_AVAILABLE, MPS_GPU_AVAILABLE
from label_sleuth.models.core.model_api import ModelAPI
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager


@dataclass
class TransformerComponents:
    model: PreTrainedModel
    language: Language


class HFTransformerModel(ModelAPI):
    """
    Basic implementation for a pytorch-based transformer model that relies on the huggingface transformers library.
    """
    def __init__(self, output_dir, background_jobs_manager: BackgroundJobsManager,
                 pretrained_model, batch_size=32, learning_rate=5e-5, num_train_epochs=5):
        """
        :param output_dir:
        :param background_jobs_manager:
        :param pretrained_model: the name of a transfomer model from huggingface.co, or a path to a directory containing
        a pytorch model created using the huggingface transformers library
        :param batch_size:
        :param learning_rate:
        :param num_train_epochs:
        """
        super().__init__(output_dir, background_jobs_manager, gpu_support=True)
        self.pretrained_model_name = pretrained_model
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.num_train_epochs = num_train_epochs
        self.max_seq_length = 128
        self.tokenizer = AutoTokenizer.from_pretrained(self.pretrained_model_name)

    def _train(self, model_id, train_data, model_params: dict):
        texts = [element["text"] for element in train_data]
        labels = [element["label"] for element in train_data]
        train_dataset = self.process_train_inputs(texts, labels)
        training_args = TrainingArguments(output_dir=self.get_models_dir(),
                                          overwrite_output_dir=True,
                                          num_train_epochs=self.num_train_epochs,
                                          per_device_train_batch_size=self.batch_size,
                                          learning_rate=self.learning_rate,
                                          use_mps_device=MPS_GPU_AVAILABLE)
        model = AutoModelForSequenceClassification.from_pretrained(self.pretrained_model_name)
        trainer = Trainer(model=model, args=training_args, train_dataset=train_dataset)
        trainer.train()
        trainer.save_model(self.get_model_dir_by_id(model_id))

    def load_model(self, model_path) -> TransformerComponents:
        language = self.get_language(model_path)
        model = AutoModelForSequenceClassification.from_pretrained(model_path)
        return TransformerComponents(model=model, language=language)

    def infer(self, model_components: TransformerComponents, items_to_infer):
        if GPU_AVAILABLE:
            if MPS_GPU_AVAILABLE:
                device = 'mps'
            else:
                device = 'cuda:0'
        else:
            device = 'cpu'

        pipeline = TextClassificationPipeline(model=model_components.model, tokenizer=self.tokenizer, device=device)

        # we sort examples by length to optimize inference time
        sorted_items_to_infer = sorted(enumerate(items_to_infer), key=lambda x: len(x[1]['text']))
        orig_idx_to_new_idx = {orig_idx: new_idx for new_idx, (orig_idx, _) in enumerate(sorted_items_to_infer)}
        ds = Dataset.from_dict({'text': [item['text'] for _, item in sorted_items_to_infer]})
        predictions = []
        for output in tqdm(pipeline(KeyDataset(ds, 'text'), batch_size=self.batch_size, truncation=True),
                           total=len(items_to_infer), desc="classification inference"):
            label = output['label'] == 'LABEL_1'
            score = output['score'] if label is True else 1-output['score']

            predictions.append(Prediction(label=label, score=score))

        predictions = [predictions[orig_idx_to_new_idx[i]] for i in range(len(predictions))]
        return predictions

    def get_model_dir_name(self):  # for backward compatibility, we override the default get_model_dir_name()
        return "transformers"

    def process_train_inputs(self, texts, labels) -> List[InputFeatures]:
        """
        Tokenize the train texts and return training data in a format expected by the transformers library.
        If the desired transformer model requires different inputs than input_ids+attention_mask+token_type_ids, this
        function may need to be overriden.
        :param texts:
        :param labels:
        :return: a list of transformers library InputFeatures
        """
        features = []
        for text, label in zip(texts, labels):
            inputs = (self.tokenizer.encode_plus(text, add_special_tokens=True, max_length=self.max_seq_length,
                                                 pad_to_max_length=True))

            features.append(InputFeatures(**inputs,
                                          label=label))
        return features


class HFBert(HFTransformerModel):
    """
    Implementation of a classifier using the BERT (Devlin et al., 2018) base model
    https://huggingface.co/bert-base-uncased
    """
    def __init__(self, output_dir, background_jobs_manager: BackgroundJobsManager):
        super().__init__(output_dir, background_jobs_manager,
                         pretrained_model="bert-base-uncased",
                         batch_size=32, learning_rate=5e-5, num_train_epochs=5)

    def get_supported_languages(self):
        return {Languages.ENGLISH}


class HFXLMRoberta(HFTransformerModel):
    """
    Implementation of a classifier using the multilingual XLM-R (Conneau et al., 2019) base model
    https://huggingface.co/xlm-roberta-base
    """
    def __init__(self, output_dir, background_jobs_manager: BackgroundJobsManager):
        super().__init__(output_dir, background_jobs_manager,
                         pretrained_model="xlm-roberta-base",
                         batch_size=16, learning_rate=2e-5, num_train_epochs=20)

    def get_supported_languages(self):
        """
        While XLM-R will not support every possible language, it should be usable (to some extent) for 100 languages,
        as specified here: https://data.statmt.org/cc-100/
        """
        return Languages.all_languages()


class HFTransformers(HFBert):
    """
    Keep old BERT class name for backward compatibility
    """



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import inspect
import logging
import threading

from dataclasses import dataclass

from label_sleuth.models.core.model_api import ModelAPI
from label_sleuth.models.core.model_type import ModelType
from label_sleuth.models.core.tools import SentenceEmbeddingService
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager


class ModelFactory:
    """
    Given a model type, this factory returns the relevant implementation of ModelAPI
    """
    def __init__(self, output_dir, background_jobs_manager: BackgroundJobsManager,
                 sentence_embedding_service: SentenceEmbeddingService):
        self.loaded_model_apis = {}
        self.model_dependencies = ModelDependencies(
            output_dir=output_dir, background_jobs_manager=background_jobs_manager,
            sentence_embedding_service=sentence_embedding_service, model_factory=self)
        self.lock = threading.RLock()

    def get_model_api(self, model_type: ModelType) -> ModelAPI:
        with self.lock:
            if model_type not in self.loaded_model_apis:
                try:
                    # check which of the model dependencies are used by this model implementation
                    model_input_args = inspect.signature(model_type.cls).parameters.keys()
                    kwargs = {k: v for k, v in self.model_dependencies.__dict__.items() if k in model_input_args}
                    # instantiate the model
                    model_api = model_type.cls(**kwargs)
                    self.loaded_model_apis[model_type] = model_api
                except Exception:
                    logging.exception(f"Could not get model type {model_type.cls} from the model factory")
        return self.loaded_model_apis[model_type]


@dataclass
class ModelDependencies:
    """
    These are the parameters that can be passed to the model's init method by the ModelFactory. Model implementations
    can use some or all of these parameters (as needed) as keyword arguments for initialization.
    """
    output_dir: str
    background_jobs_manager: BackgroundJobsManager
    sentence_embedding_service: SentenceEmbeddingService
    model_factory: ModelFactory



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import functools
import itertools
import logging
import os
import sys
import time

from collections import Counter, defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import datetime
from typing import Mapping, List, Sequence, Union, Tuple

import jsonpickle
import pandas as pd

from label_sleuth.active_learning.core.active_learning_factory import ActiveLearningFactory
from label_sleuth.analysis_utils.labeling_reports import get_suspected_labeling_contradictions_by_distance_with_diffs, \
    get_disagreements_using_cross_validation
from label_sleuth.config import Configuration
from label_sleuth.data_access.core.data_structs import DisplayFields, Document, Label, TextElement, LABEL_POSITIVE
from label_sleuth.data_access.data_access_api import DataAccessApi
from label_sleuth.data_access.label_import_utils import process_labels_dataframe
from label_sleuth.data_access.processors.csv_processor import CsvFileProcessor
from label_sleuth.definitions import ACTIVE_LEARNING_SUGGESTION_COUNT
from label_sleuth.models.core.model_api import ModelStatus
from label_sleuth.models.core.catalog import ModelsCatalog
from label_sleuth.models.core.model_type import ModelType
from label_sleuth.models.core.models_factory import ModelFactory
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.models.core.tools import SentenceEmbeddingService
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import Category, Iteration, IterationStatus, \
    ModelInfo, OrchestratorStateApi
from label_sleuth.orchestrator.utils import convert_text_elements_to_train_data
from label_sleuth.training_set_selector.training_set_selector_factory import TrainingSetSelectionFactory

# constants
NUMBER_OF_MODELS_TO_KEEP = 2
TRAIN_COUNTS_STR_KEY = "train_counts"

# members
new_data_infer_thread_pool = ThreadPoolExecutor(1)


class OrchestratorApi:
    def __init__(self, orchestrator_state: OrchestratorStateApi, data_access: DataAccessApi,
                 active_learning_factory: ActiveLearningFactory, model_factory: ModelFactory,
                 training_set_selection_factory: TrainingSetSelectionFactory,
                 background_jobs_manager: BackgroundJobsManager,
                 sentence_embedding_service: SentenceEmbeddingService, config: Configuration):
        self.orchestrator_state = orchestrator_state
        self.data_access = data_access
        self.active_learning_factory = active_learning_factory
        self.model_factory = model_factory
        self.background_jobs_manager = background_jobs_manager
        self.sentence_embedding_service = sentence_embedding_service
        self.training_set_selection_factory = training_set_selection_factory
        self.config = config
        self._verify_model_and_language_compatibility()

    def get_all_dataset_names(self):
        return sorted(self.data_access.get_all_dataset_names())

    # Workspace-related methods

    def create_workspace(self, workspace_id: str, dataset_name: str):
        """
        Create a new workspace
        :param workspace_id:
        :param dataset_name:
        """
        logging.info(f"Creating a new workspace '{workspace_id}' using dataset '{dataset_name}'")
        if dataset_name not in self.data_access.get_all_dataset_names():
            message = f"{dataset_name} does not exist. Cannot create workspace {workspace_id}"
            logging.error(message)
            raise Exception(message)
        self.orchestrator_state.create_workspace(workspace_id, dataset_name)

    def delete_workspace(self, workspace_id: str):
        """
        Delete a given workspace
        :param workspace_id:
        """
        logging.info(f"Deleting workspace '{workspace_id}'")
        if self.workspace_exists(workspace_id):
            workspace = self.orchestrator_state.get_workspace(workspace_id)
            try:
                for category_id in workspace.categories.keys():
                    self._delete_category_models(workspace_id, category_id)
                self.orchestrator_state.delete_workspace_state(workspace_id)
            except Exception as e:
                logging.exception(f"error deleting workspace '{workspace_id}'")
                raise e
            try:
                self.data_access.delete_all_labels(workspace_id, workspace.dataset_name)
            except Exception as e:
                logging.exception(f"error clearing saved labels for workspace '{workspace_id}'")
                raise e
    
    def get_workspaces_by_dataset_name(self, dataset_name: str):
        workspaces_ids = []
        for workspace_id in self.list_workspaces():
            if self.get_dataset_name(workspace_id) == dataset_name:
                workspaces_ids.append(workspace_id)
        return workspaces_ids


    def delete_dataset(self, dataset_name: str):
        """
        Delete a given workspace
        :param workspace_id:
        """
        logging.info(f"Deleting dataset '{dataset_name}'")
        

        workspaces_to_delete = self.get_workspaces_by_dataset_name(dataset_name)
        for workspace_id in workspaces_to_delete:
            self.delete_workspace(workspace_id)
        
        # delete dataset
        self.data_access.delete_dataset(dataset_name)

        res = {
            'deleted_dataset': dataset_name,
            'deleted_workspace_ids': workspaces_to_delete
        }

        return res


    def workspace_exists(self, workspace_id: str) -> bool:
        return self.orchestrator_state.workspace_exists(workspace_id)

    def list_workspaces(self):
        return sorted([x.workspace_id for x in self.orchestrator_state.get_all_workspaces()])

    def get_dataset_name(self, workspace_id: str) -> str:
        return self.orchestrator_state.get_workspace(workspace_id).dataset_name

    # Category-related methods

    def create_new_category(self, workspace_id: str, category_name: str, category_description: str):
        """
        Declare a new category in the given workspace
        :param workspace_id:
        :param category_name:
        :param category_description:st

        :return: The new category's id
        """
        logging.info(f"Creating a new category '{category_name}' in workspace '{workspace_id}'")
        return self.orchestrator_state.add_category_to_workspace(workspace_id, category_name, category_description)

    def edit_category(self, workspace_id: str, category_id: int, new_category_name: str, new_category_description: str):
        logging.info(f"Updating category id {category_id} name to {new_category_name}, and category description to "
                     f"{new_category_description}")
        return self.orchestrator_state.edit_category(workspace_id, category_id, new_category_name,
                                                     new_category_description)

    def delete_category(self, workspace_id: str, category_id: int):
        """
        Delete the given category from the workspace. This call permanently deletes all data associated with the
        category, including user labels and models.
        :param workspace_id:
        :param category_id:
        """
        logging.info(f"Deleting category id {category_id} from workspace '{workspace_id}'")
        dataset_name = self.get_dataset_name(workspace_id)
        self.data_access.delete_labels_for_category(workspace_id, dataset_name, category_id)
        self._delete_category_models(workspace_id, category_id)
        self.orchestrator_state.delete_category_from_workspace(workspace_id, category_id)

    def get_all_categories(self, workspace_id: str) -> Mapping[int, Category]:
        return self.orchestrator_state.get_all_categories(workspace_id)

    # Data access methods

    def get_documents(self, workspace_id: str, dataset_name: str, uris: Sequence[str]) -> List[Document]:
        """
        Get a list of Documents by their URIs
        :param workspace_id:
        :param dataset_name:
        :param uris:
        :return: a list of Document objects
        """
        return self.data_access.get_documents(workspace_id, dataset_name, uris)

    def get_all_document_uris(self, workspace_id) -> List[str]:
        """
        Get a list of all document URIs in the dataset used by the given workspace.
        :param workspace_id:
        :return: a list of Document URIs
        """
        dataset_name = self.get_dataset_name(workspace_id)
        return self.data_access.get_all_document_uris(dataset_name)

    def get_text_element_count(self, workspace_id) -> int:
        """
        Return the number of TextElement objects in the given dataset_name.
        :param workspace_id:
        """
        dataset_name = self.get_dataset_name(workspace_id)
        return self.data_access.get_text_element_count(dataset_name)

    def get_all_text_elements(self, dataset_name: str) -> List[TextElement]:
        """
        Get all the text elements of the given dataset.
        :param dataset_name:
        :return: a list of TextElement objects
        """
        return self.data_access.get_all_text_elements(dataset_name=dataset_name)

    def get_text_elements_by_uris(self, workspace_id: str, dataset_name: str, uris: Sequence[str]) -> List[TextElement]:
        """
        Get a list of TextElements by their URIs
        :param workspace_id:
        :param dataset_name:
        :param uris:
        :return: a list of TextElement objects
        """
        return self.data_access.get_text_elements_by_uris(workspace_id, dataset_name, uris)

    def get_all_labeled_text_elements(self, workspace_id, dataset_name, category_id: int, remove_duplicates=False) -> \
            List[TextElement]:
        """
        Get all the text elements that were assigned user labels for the given category.
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        :param remove_duplicates:
        :return: a list of TextElement objects
        """
        return self.data_access.get_labeled_text_elements(
            workspace_id, dataset_name, category_id, sample_size=sys.maxsize,
            remove_duplicates=remove_duplicates)['results']

    def get_all_unlabeled_text_elements(self, workspace_id, dataset_name, category_id: int, remove_duplicates=False) \
            -> List[TextElement]:
        """
        Get all the text elements that were not assigned user labels for the given category.
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :return: a list of TextElement objects
        """
        return self.data_access.get_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                            sample_size=sys.maxsize,
                                                            remove_duplicates=remove_duplicates)['results']

    def query(self, workspace_id: str, dataset_name: str, category_id: Union[int, None],
              query: str, is_regex: bool = False,
              sample_size: int = sys.maxsize, sample_start_idx: int = 0, unlabeled_only: bool = False,
              remove_duplicates=False) -> Mapping[str, Union[List[TextElement], int]]:
        """
        Query a dataset using the given regex, returning up to *sample_size* elements that meet the query

        :param workspace_id:
        :param dataset_name:
        :param category_id: optional. If unlabeled_only is True category_id will be used for determining unlabeled
        :param query: string
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param unlabeled_only: if True, filters out labeled elements
        :param sample_size: maximum items to return
        :param sample_start_idx: get elements starting from this index (for pagination)
        :param remove_duplicates: if True, remove duplicate elements
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """
        if unlabeled_only and category_id is None:
            raise Exception("unlabeled_only was set to True and category_id was not provided")

        if unlabeled_only:
            return self.data_access.get_unlabeled_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                                                category_id=category_id, sample_size=sample_size,
                                                                sample_start_idx=sample_start_idx,
                                                                remove_duplicates=remove_duplicates)
        else:
            return self.data_access.get_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                                      sample_size=sample_size, sample_start_idx=sample_start_idx,
                                                      query=query, is_regex=is_regex,
                                                      remove_duplicates=remove_duplicates)

    def set_labels(self, workspace_id: str, uri_to_label: Mapping[str, Mapping[int, Label]],
                   apply_to_duplicate_texts=True, update_label_counter=True):
        """
        Set user labels for a set of element URIs.
        :param workspace_id:
        :param uri_to_label: maps URIs to a dictionary in the format of {category_id: Label}, where Label is an
        instance of data_structs.Label
        :param apply_to_duplicate_texts: if True, also set the same labels for additional URIs that are duplicates of
        the URIs provided.
        :param update_label_counter: determines whether the label changes are reflected in the label change counters
        of the categories. Since an increase in label change counts can trigger the training of a new model, in some
        specific situations this parameter is set to False and the updating of the counter is performed at a later time.
        """
        if update_label_counter:
            train_set_selector = self.training_set_selection_factory.\
                get_training_set_selector(self.config.training_set_selection_strategy)

            used_label_types = train_set_selector.get_label_types()
            # count the number of labels for each category
            changes_per_cat = Counter([cat for uri, labels_dict in uri_to_label.items()
                                       for cat, cat_label in labels_dict.items()
                                       if cat_label.label_type in used_label_types])
            for cat, num_changes in changes_per_cat.items():
                self.orchestrator_state.increase_label_change_count_since_last_train(workspace_id, cat, num_changes)
        self.data_access.set_labels(workspace_id, uri_to_label, apply_to_duplicate_texts)

    def unset_labels(self, workspace_id: str, category_id: int, uris: Sequence[str], apply_to_duplicate_texts=True):
        """
        Unset labels of a set of element URIs for a given category.
        :param workspace_id:
        :param category_id:
        :param uris:
        :param apply_to_duplicate_texts: if True, also unset the same labels for additional URIs that are duplicates of
        the URIs provided.
        """
        self.data_access.unset_labels(
            workspace_id, category_id, uris, apply_to_duplicate_texts=apply_to_duplicate_texts)

    def get_label_counts(self, workspace_id: str, dataset_name: str, category_id: int, remove_duplicates=False,
                         counts_for_training=False) -> Mapping[Union[str, bool], int]:
        """
        Get the number of elements that were labeled for the given category.
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        :param remove_duplicates: whether to count all labeled elements or only unique instances
        :param counts_for_training: if True, determine the counts as relevant for training a model, e.g. lumping
        both strong and weak labels together; if False, count the different types of labels separately
        :return:
        """
        if counts_for_training:
            train_set_selector = self.training_set_selection_factory.get_training_set_selector(
                                                           self.config.training_set_selection_strategy)
            used_label_types = train_set_selector.get_label_types()
            return self.data_access.get_label_counts(workspace_id, dataset_name, category_id,
                                                     remove_duplicates=remove_duplicates,
                                                     fine_grained_counts=False,
                                                     label_types=used_label_types)
        else:
            return self.data_access.get_label_counts(workspace_id, dataset_name, category_id,
                                                     remove_duplicates=remove_duplicates, fine_grained_counts=True)

    # Iteration-related methods

    def get_all_iterations_for_category(self, workspace_id, category_id: int) -> List[Iteration]:
        """
        :param workspace_id:
        :param category_id:
        :return: dict from model_id to ModelInfo
        """
        return self.orchestrator_state.get_all_iterations(workspace_id, category_id)

    def get_all_iterations_by_status(self, workspace_id, category_id, iteration_status: IterationStatus) \
            -> List[Tuple[Iteration, int]]:
        """
        get all iterations by status
        :return A list of tuples of Iteration and iteration index
        """
        return self.orchestrator_state.get_all_iterations_by_status(workspace_id, category_id, iteration_status)

    def get_iteration_status(self, workspace_id, category_id, iteration_index) -> IterationStatus:
        return self.orchestrator_state.get_iteration_status(workspace_id, category_id, iteration_index)

    def delete_iteration_model(self, workspace_id, category_id, iteration_index):
        """
        Delete the model files for *iteration_index* of the given category, and mark the model as deleted.
        :param workspace_id:
        :param category_id:
        :param iteration_index:
        """
        iteration = self.get_all_iterations_for_category(workspace_id, category_id)[iteration_index]
        model_info = iteration.model
        if model_info.model_status == ModelStatus.DELETED:
            raise Exception(f"Trying to delete model id {model_info.model_id} which is already in {ModelStatus.DELETED}"
                            f"from workspace '{workspace_id}' in category id '{category_id}'")

        model_api = self.model_factory.get_model_api(model_info.model_type)
        logging.info(f"Marking iteration {iteration_index} model id {model_info.model_id} in "
                     f"workspace '{workspace_id}' in category id '{category_id}' as deleted, and deleting the model")
        self.orchestrator_state.mark_iteration_model_as_deleted(workspace_id, category_id, iteration_index)
        model_api.delete_model(model_info.model_id)

    def _delete_category_models(self, workspace_id, category_id):
        workspace = self.orchestrator_state.get_workspace(workspace_id)
        if workspace.categories[category_id] is not None:
            for idx in range(len(workspace.categories[category_id].iterations)):
                if workspace.categories[category_id].iterations[idx].model is not None \
                        and workspace.categories[category_id].iterations[idx].model.model_status != ModelStatus.DELETED:
                    self.delete_iteration_model(workspace_id, category_id, idx)

    def get_elements_to_label(self, workspace_id: str, category_id: int, count: int, start_index: int = 0) \
            -> Tuple[List[TextElement], int]:
        """
        Returns a list of *count* elements recommended for labeling by the active learning module for the latest
        iteration in READY status.
        :param workspace_id:
        :param category_id:
        :param count: maximum number of elements to return
        :param start_index: get elements starting from this index (for pagination)
        :return: a list of *count* TextElement objects
        """
        recommended_uris = self.orchestrator_state.get_current_category_recommendations(workspace_id, category_id)
        hit_count = len(recommended_uris)
        if start_index > len(recommended_uris):
            raise Exception(f"exceeded max recommended items. last element index is {len(recommended_uris) - 1}")
        recommended_uris = recommended_uris[start_index:start_index + count]
        dataset_name = self.get_dataset_name(workspace_id)
        return self.get_text_elements_by_uris(workspace_id, dataset_name, recommended_uris), hit_count

    # Iteration flow

    def run_iteration(self, workspace_id: str, dataset_name: str, category_id: int, model_type: ModelType):
        """
        This method initiates an Iteration, a flow that includes training a model, inferring the full corpus using
        this model, choosing candidate elements for labeling using active learning, as well as calculating various
        statistics.
        For a specific workspace and category, an iteration is identified using an integer iteration index. As different
        stages for the given iteration are completed, the IterationStatus for this iteration index is updated using the
        self.orchestrator_state.
        Since the training and inference stages of the iteration are submitted asynchronously in the background, the
        full flow is composed of this method, along with the _train_done_callback and _infer_done_callback, which are
        launched when the training and inference stages, respectively, are completed.

        :param workspace_id:
        :param dataset_name:
        :param category_id:
        :param model_type:
        """
        new_iteration_index = len(self.orchestrator_state.get_all_iterations(workspace_id, category_id))
        logging.info(f"starting iteration {new_iteration_index} in background for workspace '{workspace_id}' "
                     f"category id '{category_id}'")
        category = self.orchestrator_state.get_workspace(workspace_id).categories[category_id]
        self.orchestrator_state.add_iteration(workspace_id=workspace_id, category_id=category_id)
        train_set_selector = self.training_set_selection_factory.get_training_set_selector(
            self.config.training_set_selection_strategy)
        future = train_set_selector.collect_train_set(workspace_id=workspace_id,
                                                      train_dataset_name=dataset_name,
                                                      category_id=category_id,
                                                      category_name=category.name,
                                                      category_description=category.description)
        future.add_done_callback(functools.partial(self._train, workspace_id, category_id, model_type,
                                                   new_iteration_index))

    def _train(self, workspace_id, category_id, model_type, iteration_index, future):
        try:
            train_data = future.result()
        except Exception:
            logging.exception(f"Train set selection failed. Marking workspace '{workspace_id}' "
                              f"category id '{category_id}' iteration {iteration_index} as error")
            self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                            IterationStatus.ERROR)
            return

        def _get_counts_per_label(text_elements, category_id):
            """
            These label counts reflect the more detailed description of training labels, e.g. how many of the elements
            have weak labels
            """
            label_names = [element.category_to_label[category_id].get_detailed_label_name()
                           for element in text_elements]

            return dict(Counter(label_names))

        train_counts = _get_counts_per_label(train_data, category_id)
        train_data = convert_text_elements_to_train_data(train_data, category_id)
        train_statistics = {TRAIN_COUNTS_STR_KEY: train_counts}

        self.orchestrator_state.update_iteration_status(workspace_id=workspace_id, category_id=category_id,
                                                        iteration_index=iteration_index,
                                                        new_status=IterationStatus.TRAINING)
        model_api = self.model_factory.get_model_api(model_type)
        category = self.orchestrator_state.get_workspace(workspace_id).categories[category_id]
        logging.info(f"workspace '{workspace_id}' training a model for category id '{category_id}', "
                     f"train_statistics: {train_statistics}")
        model_params = {
            "category_id_to_info": {
                category_id: {
                    "category_name": category.name,
                    "category_description": category.description,
                }
            }
        }
        model_id, future = model_api.train(train_data=train_data, language=self.config.language,
                                           model_params=model_params)
        model_status = model_api.get_model_status(model_id)
        model_info = ModelInfo(model_id=model_id, model_status=model_status, model_type=model_type,
                               train_statistics=train_statistics, creation_date=datetime.now())
        self.orchestrator_state.add_model(workspace_id=workspace_id, category_id=category_id,
                                          iteration_index=iteration_index, model_info=model_info)
        # The train callback is added here to ensure it only runs after the iteration has been added
        future.add_done_callback(functools.partial(self._train_done_callback, workspace_id, category_id,
                                                   iteration_index))
        # The model id is returned almost immediately, but the training is performed in the background. Once training is
        # complete the iteration flow continues in the *_train_done_callback* method
        return model_id

    def _train_done_callback(self, workspace_id, category_id, iteration_index, future):
        """
        Once model training for Iteration *iteration_index* is complete, the flow of the iteration continues here. As
        part of this stage an inference job over the entire dataset is launched in the background.
        :param workspace_id:
        :param category_id:
        :param iteration_index:
        :param future: future object for the train job, which was submitted through the BackgroundJobsManager
        """
        try:
            model_id = future.result()
        except Exception:
            logging.exception(f"Train failed. Marking workspace '{workspace_id}' category id '{category_id}' "
                              f"iteration {iteration_index} as error")
            self.orchestrator_state.update_model_status(workspace_id=workspace_id, category_id=category_id,
                                                        iteration_index=iteration_index, new_status=ModelStatus.ERROR)
            self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                            IterationStatus.ERROR)
            return

        self.orchestrator_state.update_model_status(workspace_id=workspace_id, category_id=category_id,
                                                    iteration_index=iteration_index, new_status=ModelStatus.READY)
        self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                        IterationStatus.RUNNING_INFERENCE)
        iteration = self.get_all_iterations_for_category(workspace_id, category_id)[iteration_index]
        model_info = iteration.model
        model_api = self.model_factory.get_model_api(model_info.model_type)
        dataset_name = self.get_dataset_name(workspace_id)
        elements = self.get_all_text_elements(dataset_name)
        logging.info(f"Successfully trained model id {model_id} for workspace '{workspace_id}' category id "
                     f"'{category_id}' iteration {iteration_index}. Running background inference for the full "
                     f"dataset ({len(elements)} items)")
        model_api.infer_by_id_async(model_id, items_to_infer=[{"text": element.text} for element in elements],
                                    done_callback=functools.partial(self._infer_done_callback, workspace_id,
                                                                    category_id, iteration_index))
        # Inference is performed in the background. Once the infer job is complete the iteration flow continues in the
        # *_infer_done_callback* method

    def _infer_done_callback(self, workspace_id, category_id, iteration_index, future):
        """
        Once model inference for Iteration *iteration_index* over the full dataset is complete, the flow of the
        iteration continues here. As part of this stage the active learning module recommendations are calculated.
        :param workspace_id:
        :param category_id:
        :param iteration_index:
        :param future: future object for the inference job, which was submitted through the BackgroundJobsManager
        """
        try:
            predictions = future.result()
        except Exception:
            logging.exception(f"Background inference on workspace '{workspace_id}' category id '{category_id}' "
                              f"iteration {iteration_index} Failed. Marking iteration with Error")
            self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                            IterationStatus.ERROR)
            return

        try:
            logging.info(f"Successfully inferred all data for workspace_id '{workspace_id}'"
                         f" category id '{category_id}' iteration {iteration_index}, "
                         f"calculating statistics and updating active learning recommendations")

            self._calculate_iteration_statistics(workspace_id, category_id, iteration_index, predictions)

            self.orchestrator_state.update_iteration_status(workspace_id, category_id,
                                                            iteration_index, IterationStatus.RUNNING_ACTIVE_LEARNING)
            dataset_name = self.get_dataset_name(workspace_id)
            self._calculate_active_learning_recommendations(workspace_id, dataset_name, category_id,
                                                            ACTIVE_LEARNING_SUGGESTION_COUNT, iteration_index)
            self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                            IterationStatus.READY)
            logging.info(f"Successfully finished iteration {iteration_index} "
                         f"in workspace '{workspace_id}' category id '{category_id}'.")

        except Exception:
            logging.exception(f"Iteration {iteration_index} on workspace '{workspace_id}' category id '{category_id}' "
                              f"Failed. Marking iteration with Error")
            self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                            IterationStatus.ERROR)
        try:
            self._delete_old_models(workspace_id, category_id, iteration_index)
        except Exception:
            logging.exception(f"Failed to delete old models for workspace '{workspace_id}' category id '{category_id}' "
                              f"after iteration {iteration_index} finished successfully ")

    def _calculate_iteration_statistics(self, workspace_id, category_id, iteration_index,
                                        predictions: Sequence[Prediction]):
        """
        Calculate some statistics about the *iteration_index* model and store them in the workspace
        :param workspace_id:
        :param category_id:
        :param iteration_index:
        :param predictions: model predictions of the *iteration_index* model over the entire dataset
        """
        dataset_name = self.get_dataset_name(workspace_id)
        elements = self.get_all_text_elements(dataset_name)
        dataset_size = len(predictions)

        # calculate the fraction of examples that receive a positive prediction from the current model
        positive_count = sum([pred.label is True for pred in predictions])
        positive_fraction = positive_count / dataset_size
        post_train_statistics = {"positive_fraction": positive_fraction, "total_positive_count": positive_count}

        # calculate the fraction of predictions that changed between the previous model and the current model
        previous_iterations = self.orchestrator_state.get_all_iterations(workspace_id, category_id)[:iteration_index]
        previous_ready_iteration_indices = \
            [candidate_iteration_index for candidate_iteration_index, iteration in enumerate(previous_iterations)
             if iteration.status == IterationStatus.READY]
        if len(previous_ready_iteration_indices) > 0:
            previous_model_predictions = self.infer(workspace_id, category_id, elements,
                                                    iteration_index=previous_ready_iteration_indices[-1])
            num_identical = sum(x.label == y.label for x, y in zip(predictions, previous_model_predictions))
            post_train_statistics["changed_fraction"] = (dataset_size - num_identical) / dataset_size

        logging.info(f"workspace {workspace_id} category {category_id} post train measurements for "
                     f"iteration {iteration_index}: {post_train_statistics}")
        self.orchestrator_state.add_iteration_statistics(workspace_id, category_id, iteration_index,
                                                         post_train_statistics)

    def _calculate_active_learning_recommendations(self, workspace_id, dataset_name, category_id, count,
                                                   iteration_index: int):
        """
        Calculate the next recommended elements for labeling using the AL module and store them in the workspace
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        :param count:
        :param iteration_index: iteration to use
        """

        if self.config.active_learning_strategy is not None:
            active_learning_strategy = self.config.active_learning_strategy
        else:
            active_learning_strategy = self.config.active_learning_policy.get_active_learning_strategy(iteration_index)

        active_learner = self.active_learning_factory.get_active_learner(active_learning_strategy)
        logging.info(f"using active learning {active_learner}")
        # Where labels are applied to duplicate texts (the default behavior), we do not want duplicates to appear in
        # the Label Next list
        remove_duplicates = self.config.apply_labels_to_duplicate_texts
        unlabeled = self.get_all_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                         remove_duplicates=remove_duplicates)
        predictions = self.infer(workspace_id, category_id, unlabeled, iteration_index)
        new_recommendations = active_learner.get_recommended_items_for_labeling(
            workspace_id=workspace_id, dataset_name=dataset_name, category_id=category_id,
            candidate_text_elements=unlabeled, candidate_text_element_predictions=predictions, sample_size=count)
        self.orchestrator_state.update_category_recommendations(workspace_id=workspace_id, category_id=category_id,
                                                                iteration_index=iteration_index,
                                                                recommended_items=[x.uri for x in new_recommendations])

        logging.info(f"active learning recommendations for iteration index {iteration_index} "
                     f"of category id '{category_id}' are ready")

    def _delete_old_models(self, workspace_id, category_id, iteration_index):
        """
        Delete previous model files for a given workspace and category, keeping only the latest
        *NUMBER_OF_MODELS_TO_KEEP* models for which an iteration flow has completed successfully.
        :param workspace_id:
        :param category_id:
        :param iteration_index:
        """
        iterations = self.orchestrator_state.get_all_iterations(workspace_id, category_id)[:iteration_index + 1]
        ready_iteration_indices = \
            [candidate_iteration_index for candidate_iteration_index, iteration in enumerate(iterations)
             if iteration.status == IterationStatus.READY]

        for candidate_iteration_index in ready_iteration_indices[:-NUMBER_OF_MODELS_TO_KEEP]:
            logging.info(f"Deleting the model of iteration {candidate_iteration_index} as only "
                         f"{NUMBER_OF_MODELS_TO_KEEP} models are kept.")
            self.delete_iteration_model(workspace_id, category_id, candidate_iteration_index)

    def train_if_recommended(self, workspace_id: str, category_id: int, force=False) -> Union[None, str]:
        """
        Check if the minimal threshold for training a new model has been met, and if so, start the flow of a
        new Iteration.
        :param workspace_id:
        :param category_id:
        :param force: if True, launch an iteration regardless of whether the thresholds are met
        :return: None if no training was launched, else the model_id for the training job submitted in the background
        """

        workspace = self.orchestrator_state.get_workspace(workspace_id)
        dataset_name = workspace.dataset_name

        iterations = self.orchestrator_state.get_all_iterations(workspace_id, category_id).copy()

        try:
            iterations_without_errors = [iteration for iteration in iterations
                                         if iteration.status != IterationStatus.ERROR]

            changes_since_last_model = \
                self.orchestrator_state.get_label_change_count_since_last_train(workspace_id, category_id)

            label_counts = self.get_label_counts(workspace_id=workspace_id, dataset_name=dataset_name,
                                                 category_id=category_id, remove_duplicates=True,
                                                 counts_for_training=True)

            if force or (LABEL_POSITIVE in label_counts
                         and label_counts[LABEL_POSITIVE] >= self.config.first_model_positive_threshold
                         and changes_since_last_model >= self.config.changed_element_threshold):
                if len(iterations_without_errors) > 0 and iterations_without_errors[-1].status != IterationStatus.READY:
                    logging.info(f"workspace '{workspace_id}' category id '{category_id}' new elements criterion was "
                                 f"met but previous AL not yet ready, not initiating a new training")
                    return None
                self.orchestrator_state.set_label_change_count_since_last_train(workspace_id, category_id, 0)
                logging.info(
                    f"Workspace '{workspace_id}' category id '{category_id}' "
                    f"{label_counts[LABEL_POSITIVE]} positive elements (>={self.config.first_model_positive_threshold})"
                    f" {changes_since_last_model} elements changed since last model "
                    f"(>={self.config.changed_element_threshold}). Training a new model")
                iteration_num = len(iterations_without_errors)
                model_type = self.config.model_policy.get_model_type(iteration_num)
                self.run_iteration(workspace_id=workspace_id, dataset_name=dataset_name, category_id=category_id,
                                   model_type=model_type)
            else:
                logging.info(f"{label_counts[LABEL_POSITIVE]} positive elements "
                             f"(should be >={self.config.first_model_positive_threshold}) "
                             f"AND {changes_since_last_model} elements changed since last model "
                             f"(should be >={self.config.changed_element_threshold}). not training a new model")
        except Exception:
            iterations_latest = self.orchestrator_state.get_all_iterations(workspace_id, category_id)
            if len(iterations) != len(iterations_latest):
                # iteration was already created, set the status to error
                iteration_index = len(iterations_latest) - 1
                self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                                IterationStatus.ERROR)
                self.orchestrator_state.update_model_status(workspace_id=workspace_id, category_id=category_id,
                                                            iteration_index=iteration_index,
                                                            new_status=ModelStatus.ERROR)
            else:
                # iteration yet to be created, add and set status to error
                iteration_index = len(iterations_latest)
                self.orchestrator_state.add_iteration(workspace_id, category_id)
                self.orchestrator_state.update_iteration_status(workspace_id, category_id, iteration_index,
                                                                IterationStatus.ERROR)

            logging.exception(f"train_if_recommended failed in iteration {iteration_index}. Model will not be trained")

    def infer(self, workspace_id: str, category_id: int, elements_to_infer: Sequence[TextElement],
              iteration_index: int = None, use_cache: bool = True) -> Sequence[Prediction]:
        """
        Get the model predictions for a list of TextElements
        :param workspace_id:
        :param category_id:
        :param elements_to_infer: list of TextElements
        :param iteration_index: iteration to use. If set to None, the latest model for the category will be used
        :param use_cache: utilize a cache that stores inference results
        :return: a list of Prediction objects
        """
        if len(elements_to_infer) == 0:
            return []

        iterations = self.get_all_iterations_for_category(workspace_id, category_id)
        if iteration_index is None:  # use the latest ready model
            iteration = [it for it in iterations if it.status == IterationStatus.READY][-1]
        else:
            iteration = iterations[iteration_index]
            if iteration.status in [IterationStatus.PREPARING_DATA, IterationStatus.TRAINING,
                                    IterationStatus.MODEL_DELETED, IterationStatus.ERROR]:
                raise Exception(
                    f"iteration {iteration_index} in workspace '{workspace_id}' category id '{category_id}' "
                    f"is not ready for inference. "
                    f"(current status is {iteration.status}, model status is {iteration.model.model_status})")

        model_info = iteration.model
        if model_info.model_status != ModelStatus.READY:
            raise Exception(f"model id {model_info.model_id} is not in READY status "
                            f"while iteration status is {iteration.status}.  Something went wrong")

        model_api = self.model_factory.get_model_api(model_info.model_type)
        list_of_dicts = [{"text": element.text} for element in elements_to_infer]
        predictions = model_api.infer_by_id(model_id=model_info.model_id, items_to_infer=list_of_dicts,
                                            use_cache=use_cache)
        return predictions

    # Labeling/Evaluation reports

    def get_contradiction_report(self, workspace_id, category_id) -> Mapping[str, List]:
        dataset_name = self.get_dataset_name(workspace_id)
        labeled_elements = self.get_all_labeled_text_elements(workspace_id, dataset_name, category_id)
        return get_suspected_labeling_contradictions_by_distance_with_diffs(
            category_id, labeled_elements, self.sentence_embedding_service.get_sentence_embeddings_representation,
            language=self.config.language)

    def get_suspicious_elements_report(self, workspace_id, category_id,
                                       model_type: ModelType = ModelsCatalog.SVM_ENSEMBLE) -> List[TextElement]:
        dataset_name = self.get_dataset_name(workspace_id)
        labeled_elements = self.get_all_labeled_text_elements(workspace_id, dataset_name, category_id,
                                                              remove_duplicates=self.config.apply_labels_to_duplicate_texts)
        predictions = self.infer(workspace_id, category_id, labeled_elements)
        disagreements = [text_element for text_element, prediction in zip(labeled_elements, predictions)
                         if text_element.category_to_label[category_id].label != prediction.label]

        cross_validation_disagreements = get_disagreements_using_cross_validation(workspace_id, category_id,
                                                                                  labeled_elements,
                                                                                  self.model_factory,
                                                                                  self.config.language,
                                                                                  model_type)
        disagreement_uris = {text_element.uri for text_element in disagreements}
        disagreements.extend([text_element for text_element in cross_validation_disagreements if
                              text_element.uri not in disagreement_uris])
        return disagreements

    def estimate_precision(self, workspace_id, category_id, ids, changed_elements_count, iteration_index):
        dataset_name = self.get_dataset_name(workspace_id)
        text_elements = self.get_text_elements_by_uris(workspace_id, dataset_name, ids)
        positive_elements = [te for te in text_elements if te.category_to_label[category_id].label == LABEL_POSITIVE]

        estimated_precision = len(positive_elements) / len(text_elements)
        self.orchestrator_state.add_iteration_statistics(workspace_id, category_id, iteration_index,
                                                         {"estimated_precision": estimated_precision,
                                                          "estimated_precision_num_elements": len(ids)})

        # since we don't want a new model to train while labeling in precision evaluation mode, we only update the
        # labeling counts after evaluation is finished
        self.orchestrator_state.increase_label_change_count_since_last_train(workspace_id, category_id,
                                                                             changed_elements_count)
        return estimated_precision

    def increase_label_change_count_since_last_train(self, workspace_id, category_id, changed_elements_count):
        """
        Labels set in certain contexts (like precision evaluation) are not immediately reflected in the label change
        counter of the category in order to avoid triggering a new iteration. This method allows manually updating the
        label change counter.
        """
        self.orchestrator_state.increase_label_change_count_since_last_train(workspace_id, category_id,
                                                                             changed_elements_count)

    def get_elements_by_prediction(self, workspace_id, category_id, required_prediction, sample_size, start_idx=0,
                                   shuffle=False, random_state=0, remove_duplicates=True) -> List[TextElement]:
        """
        Get elements in the given workspace that received a positive prediction from the latest classification model
        for the category.
        As finding and returning _all_ the positively predicted elements is expensive, this method only collects as many
        elements as is required by the request, namely up to *sample_size + start_idx* positively predicted elements.

        :param workspace_id:
        :param category_id:
        :param required_prediction:
        :param sample_size: number of elements to return
        :param start_idx: get elements starting from this index (for pagination)
        :param shuffle: if True, text elements are retrieved in a random order.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        """

        def batched(iterable, batch_size=100):
            it = iter(iterable)
            while batch := list(itertools.islice(it, batch_size)):
                yield batch

        dataset_name = self.get_dataset_name(workspace_id)
        elements_with_required_prediction = []
        # we fetch and infer text elements in batches, and collect those elements that match the required prediction
        element_batch_iterator = batched(
            self.data_access.get_text_element_iterator(workspace_id, dataset_name, shuffle=shuffle,
                                                       random_state=random_state, remove_duplicates=remove_duplicates))
        for element_batch in element_batch_iterator:
            batch_predictions = self.infer(workspace_id, category_id, element_batch)
            for element, prediction in zip(element_batch, batch_predictions):
                if prediction.label == required_prediction:
                    elements_with_required_prediction.append(element)
            if len(elements_with_required_prediction) >= start_idx + sample_size:
                break  # we have collected enough elements
        return elements_with_required_prediction[start_idx:start_idx + sample_size]

    def get_progress(self, workspace_id: str, dataset_name: str, category_id: int):
        category_label_counts = self.get_label_counts(workspace_id, dataset_name, category_id, remove_duplicates=True,
                                                      counts_for_training=True)
        if category_label_counts[LABEL_POSITIVE]:
            changed_since_last_model_count = \
                self.orchestrator_state.get_label_change_count_since_last_train(workspace_id, category_id)

            return {"all": min(
                # for a new training to start both the number of labels changed and the number of positives must be
                # above their respective thresholds; thus, we determine the status as the minimum of the two ratios
                max(0, min(round(changed_since_last_model_count / self.config.changed_element_threshold * 100), 100)),
                max(0, min(round(category_label_counts[LABEL_POSITIVE] /
                                 self.config.first_model_positive_threshold * 100), 100)))
            }
        else:
            return {"all": 0}

    # Import/Export
    
    def import_category_labels(self, workspace_id, labels_df_to_import: pd.DataFrame):
        logging.info(f"Importing {len(labels_df_to_import)} labeled elements into workspace '{workspace_id}'"
                     f"from {len(labels_df_to_import[DisplayFields.category_name].unique())} categories")
        dataset_name = self.get_dataset_name(workspace_id)
        # Only if doc_id are provided or apply_labels_to_duplicate_texts is false the doc_id is
        # taken into account when applying labels
        apply_labels_to_duplicate_texts = self.config.apply_labels_to_duplicate_texts is True \
                                          or DisplayFields.doc_id not in labels_df_to_import.columns
        imported_categories_to_uris_and_labels, contracticting_labels_info = \
            process_labels_dataframe(workspace_id, dataset_name, self.data_access, labels_df_to_import,
                                     apply_labels_to_duplicate_texts=apply_labels_to_duplicate_texts)
        
        # name to id mapping for *existing* categories
        category_name_to_id = {category.name: category_id
                               for category_id, category in self.get_all_categories(workspace_id).items()}
        categories_counter = defaultdict(int)
        categories_created = []
        lines_skipped = []
        # if dataframe contained new categories, create them and update name to id mapping
        for category_name in imported_categories_to_uris_and_labels.keys():
            if category_name not in category_name_to_id.keys():
                logging.info(f"** category '{category_name}' is missing, creating it ** ")
                category_id = self.create_new_category(
                    workspace_id, category_name, f'{category_name} (created during upload)')
                category_name_to_id[category_name] = category_id
                categories_created.append(category_name)

        for category_name, uri_to_label in imported_categories_to_uris_and_labels.items():
            # switch from category_name to the corresponding category_id
            category_id = category_name_to_id[category_name]
            uri_to_label = {uri: {category_id: label} for uri, cat_to_label in uri_to_label.items()
                            for category_name, label in cat_to_label.items()}
            if len(uri_to_label) == 0:
                logging.info(f"found 0 elements for category {category_name}")
            else:
                logging.info(f'{category_name}: adding labels for {len(uri_to_label)} uris')
                self.set_labels(workspace_id, uri_to_label,
                                apply_to_duplicate_texts=False,
                                update_label_counter=True)

            label_counts_dict = self.get_label_counts(workspace_id, dataset_name, category_id, remove_duplicates=False)
            logging.info(f"Updated total label count in workspace '{workspace_id}' for category id {category_id} "
                         f"is {sum(label_counts_dict.values())} ({label_counts_dict})")
            categories_counter[category_id] = len(uri_to_label)
        # TODO return both positive and negative counts
        categories_counter_list = [{'category_id': key, 'counter': value} for key, value in categories_counter.items()]
        total = sum(categories_counter.values())
        
        res = {'categories': categories_counter_list,
               'categoriesCreated': categories_created,
               'linesSkipped': lines_skipped,
               'total': total,
               'contracticting_labels_info': contracticting_labels_info}
        return res

    def export_workspace_labels(self, workspace_id, labeled_only) -> pd.DataFrame:
        """
        get all user labels from the workspace as a Dataframe. Each row in the DataFrame is a label for a specific
        element for a specific category. Column names for the various fields are listed under DisplayFields.

        :param workspace_id:
        :param labeled_only: only export elements as they were labeled by the user. If set to False, use the
        TrainingSetSelectionStrategy to determine the exported elements
        """
        dataset_name = self.get_dataset_name(workspace_id)
        categories = self.get_all_categories(workspace_id)
        list_of_dicts = []

        for category_id, category in categories.items():
            label_counts = self.get_label_counts(workspace_id, dataset_name, category_id, False,
                                                 counts_for_training=True)
            total_count = sum(label_counts.values())

            if labeled_only or label_counts[LABEL_POSITIVE] == 0:  # if there are no positive elements,
                # training set selector cannot be used, so we only use the labeled elements
                logging.info(f"Labeled elements for category {category.name} ({category_id}) in workspace "
                             f"'{workspace_id}' is {total_count}")
                text_elements = self.data_access.get_labeled_text_elements(workspace_id, dataset_name, category_id,
                                                                           remove_duplicates=False)['results']
            else:
                train_set_selector = self.training_set_selection_factory.get_training_set_selector(
                    self.config.training_set_selection_strategy)
                text_elements = train_set_selector.get_train_set(workspace_id=workspace_id,
                                                                 train_dataset_name=dataset_name,
                                                                 category_id=category_id,
                                                                 category_name=category.name,
                                                                 category_description=category.description)
                logging.info(
                    f"Labeled elements size for category {category.name} ({category_id}) in workspace "
                    f"'{workspace_id}' is {total_count}, exported elements size is {len(text_elements)}")

            list_of_dicts.extend(
                [{DisplayFields.workspace_id: workspace_id,
                  DisplayFields.category_name: category.name,
                  DisplayFields.doc_id: element.uri.split('-')[1],
                  # TODO handle when handling uri/doc_id/element_id
                  DisplayFields.dataset: dataset_name,
                  DisplayFields.text: element.text,
                  DisplayFields.uri: element.uri,
                  # column for each metadata field, using the same format as in document upload
                  **{DisplayFields.csv_metadata_column_prefix + k: v for k, v in element.metadata.items()},
                  DisplayFields.label: element.category_to_label[category_id].label,
                  # TODO uncomment when adding metadata
                  # DisplayFields.label_metadata: le.category_to_label[category_id].metadata,
                  DisplayFields.label_type: element.category_to_label[category_id].label_type.name
                  }
                 for element in text_elements])

        logging.info(
            f"Exporting a total of {len(list_of_dicts)} elements from workspace '{workspace_id}'")

        return pd.DataFrame(list_of_dicts)

    def copy_model_dir_for_export(self, workspace_id, category_id, iteration_index):
        iteration = self.orchestrator_state.get_all_iterations(workspace_id, category_id)[iteration_index]

        model_api = self.model_factory.get_model_api(iteration.model.model_type)
        exported_model_dir = model_api.copy_model_dir_for_export(iteration.model.model_id)

        exported_model_info = {'model_type': iteration.model.model_type}
        model_info_encoded = jsonpickle.encode(exported_model_info)
        with open(os.path.join(exported_model_dir, 'model_info.json'), 'w') as f:
            f.write(model_info_encoded)
        return os.path.join(exported_model_dir, os.pardir)

    def add_documents_from_file(self, dataset_name, temp_file_path):
        global new_data_infer_thread_pool
        logging.info(f"adding documents to dataset '{dataset_name}'")
        documents = CsvFileProcessor(dataset_name, temp_file_path).build_documents()
        document_statistics = self.data_access.add_documents(dataset_name, documents)
        workspaces_to_update = []
        total_infer_jobs = 0
        for workspace_id in self.list_workspaces():
            if self.get_dataset_name(workspace_id) == dataset_name:
                workspaces_to_update.append(workspace_id)

                for category_id, category in self.get_all_categories(workspace_id).items():
                    if self.config.apply_labels_to_duplicate_texts:
                        # since new data may contain texts identical to existing labeled texts, we set all the existing
                        # labels again to apply the labels to the new data
                        labeled_elements = self.data_access.get_labeled_text_elements(workspace_id, dataset_name,
                                                                                      category_id)['results']
                        if len(labeled_elements) > 0:
                            uri_to_label = {te.uri: te.category_to_label for te in labeled_elements}
                            self.data_access.set_labels(workspace_id, uri_to_label, apply_to_duplicate_texts=True)

                    if len(category.iterations) > 0:
                        iteration_index = len(category.iterations) - 1
                        new_data_infer_thread_pool.submit(self._infer_missing_elements, workspace_id, category_id,
                                                          dataset_name, iteration_index)
                        total_infer_jobs += 1
        logging.info(f"done adding documents to {dataset_name} upload statistics: {document_statistics}."
                     f"{total_infer_jobs} infer jobs were submitted in the background")
        return document_statistics, workspaces_to_update

    def _infer_missing_elements(self, workspace_id, category_id, dataset_name, iteration_index):
        iteration_status = self.get_iteration_status(workspace_id, category_id, iteration_index)
        if iteration_status == IterationStatus.ERROR:
            logging.error(
                f"Cannot run inference for category id {category_id} in workspace '{workspace_id}' after new documents "
                f"were loaded to dataset '{dataset_name}' using model {iteration_index}, as the iteration status "
                f"is ERROR")
            return
        # if *model_id* is currently training, wait for training to complete
        start_time = time.time()
        while iteration_status != IterationStatus.READY:
            wait_time = time.time() - start_time
            if wait_time > 15 * 60:
                logging.error(f"Timeout reached when waiting to run inference with the last model for category id "
                              f"{category_id} in workspace '{workspace_id}' after new documents were loaded to "
                              f"dataset '{dataset_name}' using model {iteration_index}")
                return
            logging.info(f"waiting for iteration {iteration_index} to complete in order to infer newly added documents"
                         f" for category id {category_id} in workspace '{workspace_id}'")
            time.sleep(30)
            iteration_status = self.get_iteration_status(workspace_id, category_id, iteration_index)

        logging.info(f"Running inference with the latest model for category id {category_id} in workspace "
                     f"'{workspace_id}' after new documents were loaded to dataset '{dataset_name}' "
                     f"using model {iteration_index}")
        # Currently, there is no indication to the user that inference is running on the new documents, and there is no
        # indication for when this inference ends. Can be added and reflected in the UI in the future
        self.infer(workspace_id, category_id, self.get_all_text_elements(dataset_name), iteration_index)
        logging.info(f"completed inference with the latest model for category id {category_id} in workspace "
                     f"'{workspace_id}' after new documents were loaded to dataset '{dataset_name}',"
                     f"using model {iteration_index}")

    def preload_dataset(self, workspace_id):
        """

        """
        dataset_name = self.get_dataset_name(workspace_id)
        return self.data_access.preload_dataset(dataset_name)

    def _verify_model_and_language_compatibility(self):
        """
        Check that the model policy is compatible with the system language configuration
        """
        model_policy = self.config.model_policy
        for model_type in model_policy.get_all_model_types():
            model_supported_languages = self.model_factory.get_model_api(model_type).get_supported_languages()
            if self.config.language not in model_supported_languages:
                raise Exception(f"{self.config.language.name} is not supported by the model {model_type.name}, "
                                f"which is used by the configured model policy {model_policy.get_name()}")



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging

from concurrent.futures import Future
from concurrent.futures.thread import ThreadPoolExecutor

from label_sleuth.definitions import CPU_WORKERS, GPU_WORKERS, GPU_AVAILABLE


class BackgroundJobsManager:
    """
    This class manages various jobs that are submitted in the background (for example, training and inference).
    The number of jobs running on CPU/GPU at the same time is limited by the CPU_WORKERS/GPU_WORKERS parameters.
    """
    def __init__(self):
        self.cpu_executor = ThreadPoolExecutor(CPU_WORKERS, thread_name_prefix=f"CPU_{CPU_WORKERS}_threadpool")
        self.gpu_executor = ThreadPoolExecutor(GPU_WORKERS, thread_name_prefix=f"GPU_{GPU_WORKERS}_threadpool")

    def add_background_job(self, method, args, use_gpu, done_callback) -> Future:
        executor = self.get_executor(use_gpu)
        future = executor.submit(method, *args)

        logging.info(f"Adding background job {method} into the {executor._thread_name_prefix}")

        if done_callback is not None:
            future.add_done_callback(done_callback)
        return future

    def get_executor(self, model_requested_gpu) -> ThreadPoolExecutor:
        if model_requested_gpu and GPU_WORKERS > 0 and GPU_AVAILABLE:
            return self.gpu_executor
        return self.cpu_executor



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import ast
import logging
import os
import requests
import tempfile

from logging.handlers import RotatingFileHandler
from pathlib import Path
from argparse import ArgumentParser

# some operations are not yet supported by the pytorch MPS architecture, and for these operations a cpu fallback option
# needs to be specified. This environment variable must be set before importing the torch library.
os.environ['PYTORCH_ENABLE_MPS_FALLBACK'] = '1'

from label_sleuth import app
from label_sleuth.config import load_config
from label_sleuth.config import Configuration

PROJECT_ROOT = os.path.abspath(os.path.join(__file__, os.pardir))


def add_file_logger(output_path):
    """
    add a rotating file logger. Files will be written to <output_path>/logs/
    """
    log_dir = os.path.join(output_path, "logs")
    os.makedirs(log_dir, exist_ok=True)
    handler = RotatingFileHandler(
        os.path.join(log_dir, "label-sleuth.log"), maxBytes=(1048576 * 5), backupCount=7
    )

    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)


def load_sample_corpus(app, corpus_name):
    if corpus_name not in app.orchestrator_api.get_all_dataset_names():
        temp_dir = os.path.join(app.config["output_dir"], "temp", "csv_upload")
        os.makedirs(temp_dir, exist_ok=True)
        temp_file_path = os.path.join(temp_dir, f"{next(tempfile._get_candidate_names())}.csv")
        logging.info(f'Downloading sample corpus "{corpus_name}" from https://github.com/label-sleuth/data-examples/')
        response = requests.get(
            f"https://github.com/label-sleuth/data-examples/raw/main/{corpus_name}/{corpus_name}.csv")
        if response.status_code != 200:
            raise Exception(f'Corpus "{corpus_name}" could not be retrieved. Please make sure it exists in '
                            f'https://github.com/label-sleuth/data-examples/')

        with open(temp_file_path, 'wb') as f:
            f.write(response.content)
        app.orchestrator_api.add_documents_from_file(corpus_name, temp_file_path)


if __name__ == '__main__':
    parser = ArgumentParser()

    parser.add_argument('--host', type=str,
                        help='Hostname or IP address (string) on which to listen. Default is localhost',
                        default="localhost")
    parser.add_argument('--port', type=int, help='Port number. Default is 8000', default=8000)

    default_output = os.path.join(str(Path.home()), 'label-sleuth')
    parser.add_argument('--output_path', type=str, help=f'output directory. Default is your home directory: '
                                                        f'{default_output}', default=default_output)
    config_path = os.path.join(PROJECT_ROOT, "config.json")
    parser.add_argument('--config_path', type=str, help=f'Config file path. Default is {config_path}',
                        default=config_path)

    parser.add_argument('--num_serving_threads', type=int, help=f'Number of threads to use in waitress',
                        default=10)

    parser.add_argument('--load_sample_corpus', type=str,
                        help=f'Name of a sample corpus from https://github.com/label-sleuth/data-examples/, to be '
                             f'loaded at startup', default=None)

    config_args_group = parser.add_argument_group('Specific configuration parameters '
                                                  '(These override the config file specified in --config_path)')

    def process_bool_arg(bool_arg):
        return ast.literal_eval(bool_arg.title()) if type(bool_arg) == str else bool_arg

    allowed_types = [int, str, bool]
    ignore_list = ["users"]
    for attr_name, attr_type in Configuration.__annotations__.items():
        if attr_name in ignore_list:
            continue
        if attr_type in allowed_types:
            config_args_group.add_argument(f"--{attr_name}", type=process_bool_arg if attr_type == bool else attr_type,
                                           required=False)
        else:
            config_args_group.add_argument(f"--{attr_name}", type=str, required=False)

    args = parser.parse_args()
    os.makedirs(args.output_path, exist_ok=True)

    add_file_logger(args.output_path)  # log to file in addition to the console
    logging.info(f"Starting label-sleuth using config file {args.config_path}. Output directory is {args.output_path}")
    logging.info("(Starting the service for the first time may take a little longer)")
    curr_app = app.create_app(config=load_config(args.config_path, vars(args)),
                              output_dir=args.output_path)
    if args.load_sample_corpus:
        load_sample_corpus(curr_app, args.load_sample_corpus)

    app.start_server(curr_app, args.host, args.port, args.num_serving_threads)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import functools
import unittest
from unittest.mock import MagicMock

from label_sleuth.models.core.prediction import Prediction
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

DUMMY_PREDICTIONS = [Prediction(True, 0.54), Prediction(False, 0.22)]


def successful_train(mid, dummy_data):
    return mid


def train_failed_with_error(mid, dummy_data):
    raise Exception("Train failed")


def successful_infer(mid, dummy_data):
    return DUMMY_PREDICTIONS


def infer_failed_with_error(mid, dummy_data):
    raise Exception("Inference failed")


class TestBackgroundJobsManager(unittest.TestCase):
    def test_simple_training_job(self):
        callback_mock = MagicMock(name='callback')
        manager = BackgroundJobsManager()
        mid = 123
        dummy_callback_data = "workspace1"
        dummy_train_data = ["dummy"]
        future = manager.add_background_job(successful_train, (mid, dummy_train_data), False,
                                            done_callback=functools.partial(callback_mock, dummy_callback_data))
        mid_return = future.result()
        callback_mock.assert_called_once_with(dummy_callback_data, future)
        self.assertEqual(mid, mid_return)

    def test_training_error(self):
        callback_mock = MagicMock(name='callback')
        manager = BackgroundJobsManager()
        mid = 123
        dummy_callback_data = "workspace1"
        dummy_train_data = ["dummy"]
        future = manager.add_background_job(train_failed_with_error, (mid, dummy_train_data), False,
                                            done_callback=functools.partial(callback_mock, dummy_callback_data))
        self.assertRaises(Exception, future.result)
        callback_mock.assert_called_once_with(dummy_callback_data, future)

    def test_simple_infer_job(self):
        callback_mock = MagicMock(name='callback')
        manager = BackgroundJobsManager()
        mid = 123
        dummy_callback_data = "workspace1"
        dummy_train_data = ["dummy"]
        future = manager.add_background_job(successful_infer, (mid, dummy_train_data), False,
                                            done_callback=functools.partial(callback_mock, dummy_callback_data))
        predictions = future.result()
        callback_mock.assert_called_once_with(dummy_callback_data, future)
        self.assertEqual(DUMMY_PREDICTIONS, predictions)

    def test_infer_error(self):
        callback_mock = MagicMock(name='callback')
        manager = BackgroundJobsManager()
        mid = 123
        dummy_callback_data = "workspace1"
        dummy_train_data = ["dummy"]
        future = manager.add_background_job(infer_failed_with_error, (mid, dummy_train_data), False,
                                            done_callback=functools.partial(callback_mock, dummy_callback_data))
        self.assertRaises(Exception, future.result)
        callback_mock.assert_called_once_with(dummy_callback_data, future)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from label_sleuth.models.core.model_type import ModelType
from label_sleuth.models.policy.model_policy import ModelPolicy


class StaticModelPolicy(ModelPolicy):
    """
    A simple policy that is initialized using a specific classification model type and always returns this model type.
    """

    def __init__(self, model_type: ModelType = None):
        self.static_model = model_type

    def get_model_type(self, iteration_num: int) -> ModelType:
        """
        Ignores *iteration_num* and returns the model type defined in the initialization
        """
        model_type = self.static_model
        if model_type is None:
            raise ValueError('No model type was provided in the initialization')
        return model_type

    def get_all_model_types(self):
        return {self.static_model}

    def get_name(self):
        return f'Static-{self.static_model.name}'



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging

from label_sleuth.active_learning.core.active_learning_api import ActiveLearningStrategy


class ActiveLearningFactory:
    def __init__(self):
        self.active_learners = {}

    def get_active_learner(self, active_learning_strategy: ActiveLearningStrategy):
        """
        Returns an instance of an active learning module, one that implements the methods in the ActiveLearner class.
        """
        if active_learning_strategy not in self.active_learners:
            try:
                active_learner = active_learning_strategy.cls()
                self.active_learners[active_learning_strategy] = active_learner
            except Exception:
                logging.exception(f"Could not get active learning strategy {active_learning_strategy.cls} from the "
                                  f"active learning factory")

        return self.active_learners[active_learning_strategy]



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from collections import Counter

from scipy.stats import hypergeom
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_selection import mutual_info_classif

from label_sleuth.models.core.languages import Languages
from label_sleuth.models.core.tools import remove_stop_words_and_punctuation, remove_punctuation


def ngrams_by_info_gain(texts, relevant_labels, ngram_max_length, language=Languages.ENGLISH):
    cv = CountVectorizer(max_df=0.95, min_df=2, max_features=3000, ngram_range=(1, ngram_max_length),
                         stop_words=language.stop_words)
    x_vec = cv.fit_transform(texts)
    res = dict(zip(cv.get_feature_names_out(),
                   mutual_info_classif(x_vec, relevant_labels, discrete_features=True) * 100))
    return sorted(res.items(), key=lambda x: x[1], reverse=True)


def ngrams_by_hypergeometric_test(texts, relevant_labels, ngram_max_length, bonferroni_correction=False,
                                  language=Languages.ENGLISH, ngrams=None):
    positive_labels = {1, '1', True, True, True}
    positive_texts = [text for text, label in zip(texts, relevant_labels) if label in positive_labels]
    negative_texts = [text for text, label in zip(texts, relevant_labels) if label not in positive_labels]
    positive_ngrams = extract_ngrams(positive_texts, ngram_max_length, language=language)
    negative_ngrams = extract_ngrams(negative_texts, ngram_max_length, language=language)
    positive_counts = Counter(positive_ngrams)
    negative_counts = Counter(negative_ngrams)
    total_counts = positive_counts + negative_counts
    vocab = set(positive_ngrams + negative_ngrams)
    if ngrams:
        tokens_and_pvals = [(ngram, hypergeom.sf(positive_counts[ngram] - 1, sum(total_counts.values()),
                                                 sum(positive_counts.values()), total_counts[ngram]))
                            for ngram in ngrams if positive_counts[ngram] > 2]
    else:
        tokens_and_pvals = [(ngram, hypergeom.sf(positive_counts[ngram] - 1, sum(total_counts.values()),
                                                 sum(positive_counts.values()), total_counts[ngram]))
                            for ngram in vocab if positive_counts[ngram] > 2]
    if bonferroni_correction:
        # apply Bonferroni correction - multiply by number of comparisons
        tokens_and_pvals = [(ngram, p*len(tokens_and_pvals)) for ngram, p in tokens_and_pvals]
    return sorted(tokens_and_pvals, key=lambda x: x[1])


def extract_ngrams(texts, ngram_max_length, language=Languages.ENGLISH):
    unigrams = [token.lower() for text in remove_stop_words_and_punctuation(texts, language=language)
                for token in set(text.split())]
    ngrams = []
    for n in range(2, ngram_max_length+1):
        ngrams.extend(
            [' '.join(x) for text in remove_punctuation(texts)
             for x in zip(*[text.lower().split()[i:] for i in range(n)])]
        )
    return unigrams + ngrams


def get_token_overlap(text_a, text_b) -> float:
    a_tokens = set(text_a.lower().split())
    b_tokens = set(text_b.lower().split())
    intersection = a_tokens.intersection(b_tokens)
    overlap = len(intersection) / max(len(a_tokens), len(b_tokens))
    return overlap



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from label_sleuth.active_learning.core.active_learning_api import ActiveLearningStrategy
from label_sleuth.active_learning.strategies.hard_example_mining import HardMiningLearner
from label_sleuth.active_learning.strategies.random_sampling import RandomSampling
from label_sleuth.active_learning.strategies.retrospective import RetrospectiveLearner


class ActiveLearningCatalog:
    RANDOM = ActiveLearningStrategy(RandomSampling)
    HARD_MINING = ActiveLearningStrategy(HardMiningLearner)
    RETROSPECTIVE = ActiveLearningStrategy(RetrospectiveLearner)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc

from label_sleuth.active_learning.core.active_learning_api import ActiveLearningStrategy


class ActiveLearningPolicy(object, metaclass=abc.ABCMeta):
    """
    Base class for implementing an active learning policy, that determines which type of active learning strategy
    will be used. Policies can be static, i.e. always return the same active learning strategy,
    or dynamic, i.e. a different active learning strategy is returned
    depending on the current iteration.
    """

    @abc.abstractmethod
    def get_active_learning_strategy(self, iteration_num: int) -> ActiveLearningStrategy:
        """
        Given *iteration_num*, return the type of active learning strategy to be used
        :param iteration_num:
        :return: An instance of ActiveLearningStrategy
        """

    @abc.abstractmethod
    def get_name(self) -> str:
        """
        :return: a name that describes the policy
        """



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import json
import logging

from dataclasses import dataclass, field
from typing import List, Type

import dacite

from label_sleuth.active_learning.core.active_learning_api import ActiveLearningStrategy
from label_sleuth.active_learning.core.active_learning_policies import ActiveLearningPolicies
from label_sleuth.active_learning.core.catalog import ActiveLearningCatalog
from label_sleuth.active_learning.policy.active_learning_policy import ActiveLearningPolicy
from label_sleuth.models.core.languages import Language, Languages
from label_sleuth.models.core.model_policies import ModelPolicies
from label_sleuth.models.policy.model_policy import ModelPolicy
from label_sleuth.training_set_selector.train_set_selector_api import TrainSetSelectorAPI
from label_sleuth.training_set_selector.train_set_selectors import TrainSetSelectorAllLabeled
from label_sleuth.training_set_selector.train_set_selectors_catalog import TrainSetSelectorsCatalog


@dataclass
class Configuration:
    first_model_positive_threshold: int
    changed_element_threshold: int
    model_policy: ModelPolicy
    training_set_selection_strategy: Type[TrainSetSelectorAPI]
    precision_evaluation_size: int
    apply_labels_to_duplicate_texts: bool
    language: Language
    login_required: bool
    active_learning_strategy: ActiveLearningStrategy = None
    active_learning_policy: ActiveLearningPolicy = None
    main_panel_elements_per_page: int = 500
    sidebar_panel_elements_per_page: int = 50
    users: List[dict] = field(default_factory=list)


converters = {
    ModelPolicy: lambda x: getattr(ModelPolicies, x),
    Type[TrainSetSelectorAPI]: lambda x: getattr(TrainSetSelectorsCatalog, x),
    ActiveLearningStrategy: lambda x: getattr(ActiveLearningCatalog, x),
    ActiveLearningPolicy: lambda x: getattr(ActiveLearningPolicies, x),
    Language: lambda x: getattr(Languages, x)
}


def load_config(config_path, command_line_args=None) -> Configuration:
    # If this code is executed without an exception then we have a valid Configuration object
    with open(config_path) as f:
        raw_cfg = json.load(f)

    # override config with user specified values
    if command_line_args:
        for arg in command_line_args:
            if arg in raw_cfg and command_line_args[arg] is not None:
                logging.info(f"Overriding config file argument {arg} with a command line argument. "
                             f"Value changed from {raw_cfg[arg]} to {command_line_args[arg]}")
                raw_cfg[arg] = command_line_args[arg]

    config = dacite.from_dict(
        data_class=Configuration, data=raw_cfg,
        config=dacite.Config(type_hooks=converters),
    )
    if config.active_learning_strategy is None and config.active_learning_policy is None:
        raise Exception("Either active_learning_strategy or active_learning_policy must be specified")

    if config.active_learning_strategy is not None and config.active_learning_policy is not None:
        raise Exception("Only one of active_learning_strategy or active_learning_policy can be specified")

    return config



import abc
from typing import List
from label_sleuth.data_access.core.data_structs import Document


class DataProcessorAPI(object, metaclass=abc.ABCMeta):

    def get_raw_data_path(self) -> str:
        """
        Return the full path for the given dataset file.
        """
        raise NotImplementedError(f'users must define {self.get_raw_data_path.__name__} to use this base class')

    def build_documents(self) -> List[Document]:
        """
        Process the raw data into a list of Documents.

        :rtype: List[Document]
        """
        raise NotImplementedError(f'users must define {self.build_documents.__name__} to use this base class')



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import unittest
from label_sleuth.data_access.data_access_api import get_document_uri


class TestDataAccessApi(unittest.TestCase):
    def test_get_document_uri_with_date_and_time(self):
        doc_uri = get_document_uri("imported-15_Jan_2023_17-51-28-683")
        self.assertEqual("imported-15_Jan_2023_17-51-28", doc_uri)

    def test_get_document_uri(self):
        doc_uri = get_document_uri("wiki_animals_2000_pages-Bobak marmot-22")
        self.assertEqual("wiki_animals_2000_pages-Bobak marmot", doc_uri)


if __name__ == "__main__":
    unittest.main()



from label_sleuth.active_learning.core.catalog import ActiveLearningCatalog
from label_sleuth.active_learning.policy.static_active_learning_policy import StaticActiveLearningPolicy


class ActiveLearningPolicies:
    """
    Active learning policies determine which type of active learning strategy is used. Policies can be static, i.e.
    always return the same active learning strategy, or dynamic, e.g. a different strategy is returned depending on
    the current iteration number.
    """
    STATIC_RETROSPECTIVE = StaticActiveLearningPolicy(ActiveLearningCatalog.RETROSPECTIVE)
    STATIC_HARD_MINING = StaticActiveLearningPolicy(ActiveLearningCatalog.HARD_MINING)
    STATIC_RANDOM = StaticActiveLearningPolicy(ActiveLearningCatalog.RANDOM)


#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import unittest
from collections import defaultdict

from unittest.mock import MagicMock

from label_sleuth.analysis_utils.labeling_reports import get_suspected_labeling_contradictions_by_distance
from label_sleuth.data_access.core.data_structs import TextElement, Label, LabelType
from label_sleuth.models.core.languages import Languages


class TestLabelingReports(unittest.TestCase):

    def setUp(self) -> None:
        self.embedding_func_mock = MagicMock()

        def side_effect(texts, language):
            text_to_embedding = {"almost identical text 1": [0.9, 0.9, 0.9],
                                 "almost identical text 2": [1, 1, 0.9],
                                 "almost identical text 3": [1, 0.9, 1],
                                 "almost identical text 4": [1, 1, 1]}
            return [text_to_embedding.get(text, [0, 0, 0]) for text in texts]

        self.embedding_func_mock.side_effect = side_effect

    def test_one_positive_one_negative(self):
        labeled_elements = [TextElement(uri="dataset-uri1",
                                        text='almost identical text 1',
                                        span=[(217, 377)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=True, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )})),
                            TextElement(uri="dataset-uri2",
                                        text='almost identical text 3',
                                        span=[(108, 216)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=False, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )}))]

        suspected_contradictions = get_suspected_labeling_contradictions_by_distance(3, labeled_elements,
                                                                                     self.embedding_func_mock,
                                                                                     Languages.ENGLISH)
        self.assertEqual(len(suspected_contradictions), 1)
        self.assertEqual(suspected_contradictions[0][0].text, "almost identical text 1")
        self.assertEqual(suspected_contradictions[0][1].text, "almost identical text 3")

    def test_one_positive_two_negatives(self):
        labeled_elements = [TextElement(uri="dataset-uri1",
                                        text='almost identical text 1',
                                        span=[(217, 377)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=True, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )})),
                            TextElement(uri="dataset-uri2",
                                        text='almost identical text 2',
                                        span=[(108, 216)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=False, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )})),
                            TextElement(uri="dataset-uri3",
                                        text='almost identical text 3',
                                        span=[(108, 216)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=False, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )}))]

        suspected_contradictions = get_suspected_labeling_contradictions_by_distance(3, labeled_elements,
                                                                                     self.embedding_func_mock,
                                                                                     Languages.ENGLISH)
        self.assertEqual(len(suspected_contradictions), 2)
        self.assertEqual(suspected_contradictions[0][0].text, "almost identical text 1")
        self.assertEqual(suspected_contradictions[0][1].text, "almost identical text 2")
        self.assertEqual(suspected_contradictions[1][0].text, "almost identical text 1")
        self.assertEqual(suspected_contradictions[1][1].text, "almost identical text 3")

    def test_two_positives_two_negatives(self):
        labeled_elements = [TextElement(uri="dataset-uri1",
                                        text='almost identical text 1',
                                        span=[(217, 377)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=True, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )})),
                            TextElement(uri="dataset-uri2",
                                        text='almost identical text 2',
                                        span=[(217, 377)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=True, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )}))
            , TextElement(uri="dataset-uri3",
                          text='almost identical text 3',
                          span=[(108, 216)], metadata={},
                          category_to_label=defaultdict(Label, {3: Label(label=False, metadata={},
                                                                         label_type=LabelType.Standard

                                                                         )})),
                            TextElement(uri="dataset-uri4",
                                        text='almost identical text 4',
                                        span=[(108, 216)], metadata={},
                                        category_to_label=defaultdict(Label, {3: Label(label=False, metadata={},
                                                                                       label_type=LabelType.Standard

                                                                                       )}))]

        suspected_contradictions = get_suspected_labeling_contradictions_by_distance(3, labeled_elements,
                                                                                     self.embedding_func_mock,
                                                                                     Languages.ENGLISH)
        self.assertEqual(len(suspected_contradictions), 2)
        self.assertEqual(suspected_contradictions[0][0].text, "almost identical text 2")
        self.assertEqual(suspected_contradictions[0][1].text, "almost identical text 4")
        self.assertEqual(suspected_contradictions[1][0].text, "almost identical text 1")
        self.assertEqual(suspected_contradictions[1][1].text, "almost identical text 3")



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc
import logging
import os
import shutil
import threading
import uuid
import tempfile

from collections import defaultdict
from concurrent.futures import Future
from enum import Enum
from typing import Mapping, Sequence, Tuple, Set

import jsonpickle

import label_sleuth.definitions as definitions
from label_sleuth.models.core.languages import Languages, Language
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.models.util.disk_cache import load_model_prediction_store_from_disk, \
    save_model_prediction_store_to_disk
from label_sleuth.models.util.LRUCache import LRUCache
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

PREDICTIONS_STORE_DIR_NAME = "predictions"
LANGUAGE_STR_KEY = "Language"


class ModelStatus(Enum):
    TRAINING = 0
    READY = 1
    ERROR = 2
    DELETED = 3


class ModelAPI(object, metaclass=abc.ABCMeta):
    """
    Base class for implementing a classification model.
    This base class provides general methods for training in the background, caching model predictions etc.,
    while the _train(), load_model() and infer() methods are specific to each model implementation.
    """
    def __init__(self, output_dir, background_jobs_manager: BackgroundJobsManager, gpu_support=False):
        """
        Model implementations can require some or all of the parameters in models_factory.ModelDependencies
        in their __init__ method, as these will be passed to the model by the ModelFactory.
        Only parameters from ModelDependencies can be required by the __init__ of a new ModelAPI implementation.
        """
        self.output_dir = output_dir
        os.makedirs(self.get_models_dir(), exist_ok=True)
        self.background_jobs_manager = background_jobs_manager
        self.gpu_support = gpu_support
        self.primary_lock = threading.Lock()
        self.model_locks = defaultdict(lambda: threading.Lock())
        self.cache = LRUCache(definitions.INFER_CACHE_SIZE)
        self.cache_lock = threading.Lock()

    @abc.abstractmethod
    def _train(self, model_id: str, train_data: Sequence[Mapping], model_params: Mapping):
        """
        Method for training a classification model on *train_data*. This method is specific to each classification
        model, and is typically launched in the background via the train() method.
        :param model_id: a unique id for the model (which was generated by the train() method)
        :param train_data: a list of dictionaries with at least the "text" and "label" fields, additional fields can be
        passed e.g. [{'text': 'text1', 'label': True, 'additional_field': 'value1'}, {'text': 'text2', 'label': False,
        'additional_field': 'value2'}]
        :param model_params: dictionary for additional model parameters (can be None)
        """

    @abc.abstractmethod
    def load_model(self, model_path: str):
        """
        Return an object for the model in *model_path*, that contains all the components that are necessary to
        perform inference (e.g., the trained model itself, the language recognized by the model, a trained
        vectorizer/tokenizer etc.).
        This method is specific to each classification model, and each model implementation can return a different
        object with different components.
        :param model_path: path to a folder containing the model
        """

    @abc.abstractmethod
    def infer(self, model_components, items_to_infer) -> Sequence[Prediction]:
        """
        Perform inference using *model_components* on *items_to_infer*, and return the predictions. This method is
        specific to each classification model.
        :param model_components: an object containing all the components that are necessary to perform inference.
        :param items_to_infer: a list of dictionaries with at least the "text" field, additional fields can be passed
        e.g. [{'text': 'text1', 'additional_field': 'value1'}, {'text': 'text2', 'additional_field': 'value2'}]
        :return: a list of Prediction objects - one for each item in *items_to_infer* - where Prediction.label is a
        boolean and Prediction.score is a float in the range [0-1].
        If the model returns a non-standard type of prediction object - i.e. one that inherits from the base Prediction
        class and adds additional outputs - it must override the get_predictions_class() method.
        """

    @abc.abstractmethod
    def get_supported_languages(self) -> Set[Language]:
        """
        Each model implementation must specify which language(s) it supports
        """

    def get_prediction_class(self):
        """
        Returns the prediction dataclass used by the model. This class is used for storing and loading model
        predictions from the disk.
        """
        return Prediction

    def train(self, train_data: Sequence[Mapping], language: Language,
              model_params=None, done_callback=None) -> Tuple[str, Future]:
        """
        Create a unique model identifier, and launch a model training job in a background thread.
        :param train_data: a list of dictionaries with at least the "text" and "label" fields, additional fields can be
        passed e.g. [{'text': 'text1', 'label': True, 'additional_field': 'value1'}, {'text': 'text2', 'label': False,
        'additional_field': 'value2'}]
        :param language: the language used to initialize the model. The implemented _train() and load_model() methods
        can then access this parameter via the get_language() call
        :param model_params: dictionary for additional model parameters (can be None)
        :param done_callback: an optional function to be executed once the training job has completed
        :return: a unique identifier for the model, and a Future object for the training job that was submitted in the
        background
        """
        if model_params is None:
            model_params = {}
        model_id = f"{self.__class__.__name__}_{str(uuid.uuid1())}"
        self.mark_train_as_started(model_id)
        self.save_metadata(model_id, language, model_params)

        future = self.background_jobs_manager.add_background_job(self.train_and_update_status,
                                                                 args=(model_id, train_data, model_params),
                                                                 use_gpu=self.gpu_support,
                                                                 done_callback=done_callback)
        return model_id, future

    def train_and_update_status(self, model_id, *args) -> str:
        """
        Run the model _train() function, and return *model_id* if it has finished successfully.
        """
        try:
            self._train(model_id, *args)
            self.mark_train_as_completed(model_id)
        except Exception:
            logging.exception(f'model {model_id} failed with exception')
            self.mark_train_as_error(model_id)
            raise

        return model_id

    @staticmethod
    def _infer_item_to_cache_key(item:Mapping):
        """
        returns the unique identifier of an item sent to inference. Currently an item consists only of a text field.
        * The return value is somewhat complex for backward compatability. In the next breaking change, this could
        be changed to items's text only
        """
        return str(tuple(sorted(item.items())))

    def infer_by_id(self, model_id, items_to_infer: Sequence[Mapping], use_cache=True) -> Sequence[Prediction]:
        """
        Infer using *model_id* on *items_to_infer*, and return the predictions. This method wraps the _infer_by_id()
        method which performs the inference itself, adding prediction caching functionality with both an in-memory
        cache and a prediction store on disk.
        Thus, if *use_cache* is True, inference is only performed on items for which there are no predictions in either
        the cache or the disk store, and all predictions for *items_to_infer* are saved to both.

        :param model_id:
        :param items_to_infer: a list of dictionaries with at least the "text" field, additional fields can be passed
        e.g. [{'text': 'text1', 'additional_field': 'value1'}, {'text': 'text2', 'additional_field': 'value2'}]
        :param use_cache: determines whether to use the caching functionality. Default is True
        :return: a list of Prediction objects, one for each item in *items_to_infer*
        """
        if not use_cache:
            logging.info(f"Running infer without cache for {len(items_to_infer)} values in {self.__class__.__name__} "
                         f"model id {model_id}")
            return self._infer_by_id(model_id, items_to_infer)

        in_memory_cache_keys = [(model_id, self._infer_item_to_cache_key(item)) for item in items_to_infer]

        # If there are multiple calls to infer_by_id() using the same *model_id*, we do not want them to perform the
        # below logic at the same time. Specifically, if two calls are asking for prediction results for the same
        # element, the desired behavior is that just one of the calls will perform inference (if necessary) and save the
        # prediction results to the cache; after that, the other call can retrieve the results directly from the cache.
        # Thus, we use a lock per model_id.
        model_lock_object = self._get_model_lock_object(model_id)
        with model_lock_object:
            # Try to get the predictions from the in-memory cache.
            with self.cache_lock:  # we avoid different threads reading and writing to the cache at the same time
                infer_res = [self.cache.get(cache_key) for cache_key in in_memory_cache_keys]

            indices_not_in_cache = [i for i, v in enumerate(infer_res) if v is None]

            if len(indices_not_in_cache) > 0:  # i.e., some items aren't in the in-memory cache
                logging.info(f"{len(indices_not_in_cache)} not in cache, loading model prediction store from disk "
                             f"in {self.__class__.__name__} for model {model_id}")
                model_predictions_store = self._load_model_prediction_store_to_cache(model_id)
                logging.info(f"done loading model prediction store from disk in {self.__class__.__name__} for "
                             f"model id {model_id}")
                for idx in indices_not_in_cache:
                    infer_res[idx] = self.cache.get(in_memory_cache_keys[idx])
                indices_not_in_cache = [i for i, v in enumerate(infer_res) if v is None]

            if len(indices_not_in_cache) > 0:  # i.e., some items aren't in the in-memory cache or the prediction store
                logging.info(f"model id {model_id}, {len(items_to_infer) - len(indices_not_in_cache)} already in cache, running inference "
                             f"for {len(indices_not_in_cache)} values (cache size {self.cache.get_current_size()}) "
                             f"in {self.__class__.__name__}")
                missing_items_to_infer = [items_to_infer[idx] for idx in indices_not_in_cache]
                # If duplicates exist, do not infer the same item more than once
                uniques = set()
                uniques_to_infer = [e for e in missing_items_to_infer if frozenset(e.items()) not in uniques
                                    and not uniques.add(frozenset(e.items()))]

                # Run inference using the model for the missing elements
                new_predictions = self._infer_by_id(model_id, uniques_to_infer)
                logging.info(f"finished running infer for {len(indices_not_in_cache)} values")

                item_to_prediction = {frozenset(unique_item.items()): item_predictions
                                      for unique_item, item_predictions in zip(uniques_to_infer, new_predictions)}

                # each model has a separate store so model id should not be part of the cache key
                model_predictions_store_keys = [cache_key[1] for cache_key in in_memory_cache_keys]

                # Update cache and prediction store with predictions for the newly inferred elements
                with self.cache_lock:
                    for idx, entry in zip(indices_not_in_cache, missing_items_to_infer):
                        prediction = item_to_prediction[frozenset(entry.items())]
                        infer_res[idx] = prediction
                        self.cache.set(in_memory_cache_keys[idx], prediction)
                        model_predictions_store[model_predictions_store_keys[idx]] = prediction
                save_model_prediction_store_to_disk(self.get_model_prediction_store_file(model_id),
                                                    model_predictions_store)
            return infer_res

    def infer_by_id_async(self, model_id, items_to_infer: Sequence[Mapping], done_callback=None):
        """
        Used for launching an inference job in the background. This method has no return, and is suited for a situation
        where the goal is to run a (potentially) long inference job and cache the results. After this background
        inference is complete, calls to infer_by_id() with *model_id* can fetch the prediction results from the cache.

        :param model_id:
        :param items_to_infer: a list of dictionaries with at least the "text" field, additional fields can be passed
        e.g. [{'text': 'text1', 'additional_field': 'value1'}, {'text': 'text2', 'additional_field': 'value2'}]
        :param done_callback: an optional function to be executed once the inference job has completed
        """
        self.background_jobs_manager.add_background_job(self.infer_by_id, args=(model_id, items_to_infer),
                                                        use_gpu=self.gpu_support, done_callback=done_callback)

    def _infer_by_id(self, model_id, items_to_infer):
        model_components = self.load_model(model_path=self.get_model_dir_by_id(model_id))
        return self.infer(model_components, items_to_infer)

    def get_model_status(self, model_id) -> ModelStatus:
        if os.path.isfile(self.get_completed_flag_path(model_id)):
            return ModelStatus.READY
        elif os.path.isfile(self.get_in_progress_flag_path(model_id)):
            return ModelStatus.TRAINING
        return ModelStatus.ERROR

    def delete_model(self, model_id):
        logging.info(f"Deleting {self.__class__.__name__} model id {model_id}")
        model_dir = self.get_model_dir_by_id(model_id)
        if os.path.isdir(model_dir):
            shutil.rmtree(model_dir)
        prediction_store_path = self.get_model_prediction_store_file(model_id)
        if os.path.exists(prediction_store_path):
            logging.info(f"Deleting prediction store {prediction_store_path}")
            os.remove(prediction_store_path)

    def mark_train_as_started(self, model_id):
        os.makedirs(self.get_model_dir_by_id(model_id), exist_ok=True)
        with open(self.get_in_progress_flag_path(model_id), 'w') as f:
            pass

    def mark_train_as_completed(self, model_id):
        with open(self.get_completed_flag_path(model_id), 'w') as f:
            pass
        os.remove(self.get_in_progress_flag_path(model_id))

    def mark_train_as_error(self, model_id):
        os.remove(self.get_in_progress_flag_path(model_id))

    def get_completed_flag_path(self, model_id):
        return os.path.join(self.get_model_dir_by_id(model_id), f'train_complete')

    def get_in_progress_flag_path(self, model_id):
        return os.path.join(self.get_model_dir_by_id(model_id), f'train_in_progress')

    def save_metadata(self, model_id, language: Language, model_params: Mapping):
        """
        Saves metadata on the model training parameters (e.g. model language) to disk.
        Specifically, this metadata dictionary should include values for the keys in METADATA_PARAMS_AND_DEFAULTS
        """
        metadata_path = os.path.join(self.get_model_dir_by_id(model_id), 'model_metadata.json')
        model_metadata = {LANGUAGE_STR_KEY: language.name, **model_params}

        with open(metadata_path, 'w') as f:
            f.write(jsonpickle.encode(model_metadata))

    @staticmethod
    def get_metadata(model_path: str):
        metadata_path = os.path.join(model_path, 'model_metadata.json')
        with open(metadata_path, 'r') as f:
            metadata = jsonpickle.decode(f.read())
        return metadata

    @staticmethod
    def get_language(model_path: str) -> Language:
        language_name = ModelAPI.get_metadata(model_path)[LANGUAGE_STR_KEY]
        return getattr(Languages, language_name.upper())

    def _load_model_prediction_store_to_cache(self, model_id):
        logging.debug(f"start loading cache from disk from {self.get_model_prediction_store_file(model_id)}")
        model_predictions_store = load_model_prediction_store_from_disk(self.get_model_prediction_store_file(model_id),
                                                                        self.get_prediction_class())
        logging.debug(f"done loading cache from disk from {self.get_model_prediction_store_file(model_id)}")
        for key, value in model_predictions_store.items():
            self.cache.set((model_id, key), value)
        return model_predictions_store

    def get_model_prediction_store_file(self, model_id):
        return os.path.join(self.get_models_dir(), PREDICTIONS_STORE_DIR_NAME, model_id + ".json")

    def get_model_dir_by_id(self, model_id):
        return os.path.join(self.get_models_dir(), model_id)

    def get_models_dir(self) -> str:
        """
        Returns the base output directory for saving the models and predictions cache.
        """
        return os.path.join(self.output_dir, self.get_model_dir_name())

    def get_model_dir_name(self):
        """
        The name of the directory inside self.output_dir in which models will be saved
        """
        return self.__class__.__name__

    def copy_model_dir_for_export(self, model_id):
        model_path = self.get_model_dir_by_id(model_id)
        temp_path = str(tempfile.mkdtemp())
        output_path = os.path.join(temp_path, os.path.basename(model_path))
        shutil.copytree(model_path, output_path)

        return output_path

    def _get_model_lock_object(self, model_id):
        """
        Get a lock for a model. self.primary_lock is necessary because self.model_locks is not thread safe
        """
        with self.primary_lock:
            return self.model_locks[model_id]



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
import threading
import logging
from collections import defaultdict

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Dict, List, Sequence, Tuple, Mapping

import jsonpickle

from label_sleuth.models.core.model_api import ModelStatus
from label_sleuth.models.core.model_type import ModelType


class IterationStatus(Enum):
    PREPARING_DATA = -1  # negative numbering is due to backward compatibility issues
    TRAINING = 0
    RUNNING_INFERENCE = 1
    RUNNING_ACTIVE_LEARNING = 2
    CALCULATING_STATISTICS = 3
    READY = 4
    ERROR = 5
    MODEL_DELETED = 6


@dataclass
class ModelInfo:
    model_id: str
    model_status: ModelStatus
    creation_date: datetime
    model_type: ModelType
    train_statistics: dict


@dataclass
class Iteration:
    """
    This class stores information about an iteration. The flow of an iteration includes training a model,
    inferring the full corpus using this model, choosing candidate elements for labeling using active learning, as
    well as calculating various statistics. Each iteration is associated with a specific Category.
    """
    status: IterationStatus
    model: ModelInfo = None
    iteration_statistics: Dict = field(default_factory=dict)
    active_learning_recommendations: Sequence[str] = field(default_factory=list)


@dataclass
class Category:
    name: str
    id: int
    description: str
    label_change_count_since_last_train: int = 0
    iterations: List[Iteration] = field(default_factory=list)


@dataclass
class Workspace:
    workspace_id: str
    dataset_name: str
    categories: Dict[int, Category] = field(default_factory=dict)


class WorkspaceSchemeChangedException(Exception):
    def __init__(self, message):
        self.message = message


class OrchestratorStateApi:

    def __init__(self, workspaces_dir):
        self.workspace_dir = workspaces_dir
        os.makedirs(self.workspace_dir, exist_ok=True)
        self.workspaces = dict()  # in-memory cache for Workspace objects
        self.workspaces_lock = defaultdict(threading.RLock)  # lock for methods that access or manipulate the workspaces

    # Workspace-related methods

    def create_workspace(self, workspace_id: str, dataset_name: str):
        with self.workspaces_lock[workspace_id]:
            illegal_chars = "".join(x for x in workspace_id if not x.isalnum() and x not in "_-")
            assert len(illegal_chars) == 0, \
                f"Workspace id '{workspace_id}' contains illegal characters: '{illegal_chars}'"

            workspace = Workspace(workspace_id=workspace_id, dataset_name=dataset_name)

            if self._filename_from_workspace_id(workspace_id) in os.listdir(self.workspace_dir):
                raise Exception(f"workspace name '{workspace_id}' already exists")
            self._save_workspace(workspace)

    def get_workspace(self, workspace_id) -> Workspace:
        with self.workspaces_lock[workspace_id]:
            try:
                return self._load_workspace(workspace_id)
            except Exception as e:
                raise WorkspaceSchemeChangedException(
                    f"Failed to load workspace {workspace_id}. This is probably due to an unhandled "
                    f"workspace scheme upgrade. Please open an issue on "
                    f"https://github.com/label-sleuth/label-sleuth/issues/new/choose for additional support") from e

    def get_all_categories(self, workspace_id) -> Mapping[int, Category]:
        return {category_id: category for category_id, category in
                self.get_workspace(workspace_id).categories.items() if category is not None}

    def workspace_exists(self, workspace_id: str) -> bool:
        with self.workspaces_lock[workspace_id]:
            return os.path.exists(os.path.join(self.workspace_dir, self._filename_from_workspace_id(workspace_id)))

    def delete_workspace_state(self, workspace_id: str):
        with self.workspaces_lock[workspace_id]:
            os.remove(os.path.join(self.workspace_dir, self._filename_from_workspace_id(workspace_id)))
            if workspace_id in self.workspaces:
                del self.workspaces[workspace_id]

    def get_all_workspaces(self) -> Sequence[Workspace]:
        all_workspaces = []
        for file in os.listdir(self.workspace_dir):
            if file.startswith('.') or not file.endswith('.json'):
                continue
            workspace_id, _ = os.path.splitext(file)
            try:
                workspace = self.get_workspace(workspace_id)
            except WorkspaceSchemeChangedException:
                logging.warning(f"Failed to load workspace {workspace_id}. skipping...", exc_info=True)
                continue
            all_workspaces.append(workspace)
        return all_workspaces

    def _load_workspace(self, workspace_id) -> Workspace:
        cached_workspace = self.workspaces.get(workspace_id)
        if cached_workspace:
            return cached_workspace
        with open(os.path.join(self.workspace_dir, self._filename_from_workspace_id(workspace_id))) as json_file:
            workspace = json_file.read()
        workspace = jsonpickle.decode(workspace)
        # int dictionary keys are converted to string when writing to json, see https://bugs.python.org/issue34972
        workspace.categories = {int(category_id_str): category
                                for category_id_str, category in workspace.categories.items()}
        self.workspaces[workspace_id] = workspace
        return workspace

    def _save_workspace(self, workspace: Workspace):
        workspace_encoded = jsonpickle.encode(workspace)
        with open(os.path.join(self.workspace_dir, self._filename_from_workspace_id(workspace.workspace_id)), 'w') as f:
            f.write(workspace_encoded)

    @staticmethod
    def _filename_from_workspace_id(workspace_id: str):
        return workspace_id + ".json"

    # Category-related methods

    def add_category_to_workspace(self, workspace_id: str, category_name: str, category_description: str):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            existing_category_names = [category.name for category in workspace.categories.values()
                                       if category is not None]
            if category_name in existing_category_names:
                raise Exception(f"Category '{category_name}' already exists in workspace '{workspace_id}'")
            category_id = len(workspace.categories)
            workspace.categories[category_id] = Category(name=category_name, description=category_description,
                                                         id=category_id)
            self._save_workspace(workspace)
            return category_id

    def edit_category(self, workspace_id: str, category_id: int, new_category_name: str, new_category_description: str):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            workspace.categories[category_id].name = new_category_name
            workspace.categories[category_id].description = new_category_description

            self._save_workspace(workspace)
        return category_id

    def delete_category_from_workspace(self, workspace_id: str, category_id: int):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            workspace.categories[category_id] = None
            self._save_workspace(workspace)

    def get_current_category_recommendations(self, workspace_id: str, category_id: int) -> Sequence[str]:
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            category = workspace.categories[category_id]

            for iteration in reversed(category.iterations):
                if iteration.status == IterationStatus.READY:
                    return iteration.active_learning_recommendations
            return []

    def update_category_recommendations(self, workspace_id: str, category_id: int, iteration_index: int,
                                        recommended_items: Sequence[str]):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            workspace.categories[category_id].iterations[iteration_index].active_learning_recommendations \
                = recommended_items
            self._save_workspace(workspace)

    def get_label_change_count_since_last_train(self, workspace_id: str, category_id: int) -> int:
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            return workspace.categories[category_id].label_change_count_since_last_train

    def set_label_change_count_since_last_train(self, workspace_id: str, category_id: int, number_of_changes: int):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            workspace.categories[category_id].label_change_count_since_last_train = number_of_changes
            self._save_workspace(workspace)

    def increase_label_change_count_since_last_train(self, workspace_id: str, category_id: int,
                                                     number_of_new_changes: int):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            workspace.categories[category_id].label_change_count_since_last_train = \
                workspace.categories[category_id].label_change_count_since_last_train + number_of_new_changes
            self._save_workspace(workspace)

    # Iteration-related methods

    def add_iteration(self, workspace_id: str, category_id: int):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            iteration = Iteration(status=IterationStatus.PREPARING_DATA)
            workspace.categories[category_id].iterations.append(iteration)
            self._save_workspace(workspace)

    def add_model(self, workspace_id: str, category_id: int, iteration_index: int, model_info: ModelInfo):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            if workspace.categories[category_id].iterations[iteration_index].model is not None:
                raise Exception(f"Workspace '{workspace_id}' category '{category_id}' iteration {iteration_index} "
                                f"already has a model, cannot add a model")
            workspace.categories[category_id].iterations[iteration_index].model = model_info
            self._save_workspace(workspace)

    def get_iteration_status(self, workspace_id: str, category_id: int, iteration_index: int) -> IterationStatus:
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            return workspace.categories[category_id].iterations[iteration_index].status

    def update_iteration_status(self, workspace_id: str, category_id: int, iteration_index: int,
                                new_status: IterationStatus):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            workspace.categories[category_id].iterations[iteration_index].status = new_status
            self._save_workspace(workspace)

    def get_all_iterations(self, workspace_id, category_id: int) -> List[Iteration]:
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            return workspace.categories[category_id].iterations

    def get_all_iterations_by_status(self, workspace_id: str, category_id: int, status: IterationStatus) -> \
            List[Tuple[Iteration, int]]:
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            return [(iteration, idx) for idx, iteration in enumerate(workspace.categories[category_id].iterations)
                    if iteration.status == status]

    def add_iteration_statistics(self, workspace_id, category_id: int, iteration_index: int, statistics_dict: dict):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            iteration = workspace.categories[category_id].iterations[iteration_index]
            iteration.iteration_statistics.update(statistics_dict)
            self._save_workspace(workspace)

    def update_model_status(self, workspace_id: str, category_id: int, iteration_index: int, new_status: ModelStatus):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            iterations = workspace.categories[category_id].iterations
            assert len(iterations) > iteration_index,\
                f"Iteration '{iteration_index}' doesn't exist in workspace '{workspace_id}'"
            iterations[iteration_index].model.model_status = new_status
            self._save_workspace(workspace)

    def mark_iteration_model_as_deleted(self, workspace_id, category_id: int, iteration_index: int):
        with self.workspaces_lock[workspace_id]:
            workspace = self._load_workspace(workspace_id)
            iteration = self.get_all_iterations(workspace_id, category_id)[iteration_index]
            iteration.model.model_status = ModelStatus.DELETED
            iteration.status = IterationStatus.MODEL_DELETED
            self._save_workspace(workspace)
